/*
 * Electric(tm) VLSI Design System
 *
 * File: erc.c
 * Electrical Rules Checking tool
 * Written by: Steven M. Rubin
 *
 * Copyright (c) 2000 Static Free Software.
 *
 * Electric(tm) is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Electric(tm) is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Electric(tm); see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, Mass 02111-1307, USA.
 *
 * Static Free Software
 * 4119 Alpine Road
 * Portola Valley, California 94028
 * info@staticfreesoft.com
 */

#include "config.h"
#if ERCAID

#include "global.h"
#include "efunction.h"
#include "egraphics.h"
#include "erc.h"
#include "edialogs.h"

/* the ERC aid table */
static KEYWORD ercopt[] =
{
	{"analyze",        0,{NOKEY,NOKEY,NOKEY,NOKEY,NOKEY}},
	TERMKEY
};
COMCOMP erc_tablep = {ercopt, NOTOPLIST, NONEXTLIST, NOPARAMS, NOBACKUP,
	0, " \t", M_("Electrical Rules action"), M_("show defaults")};

/* well contacts */
#define NOWELLCON ((WELLCON *)-1)

typedef struct Iwellcon
{
	INTBIG           x, y;
	INTBIG           netnum;
	INTBIG           fun;
	NODEPROTO       *np;
	struct Iwellcon *nextwellcon;
} WELLCON;

WELLCON    *erc_wellconfree = NOWELLCON;
WELLCON    *erc_firstwellcon = NOWELLCON;
INTBIG      erc_worstpwelldist;
INTBIG      erc_worstpwellconx, erc_worstpwellcony;
INTBIG      erc_worstpwelledgex, erc_worstpwelledgey;
INTBIG      erc_worstnwelldist;
INTBIG      erc_worstnwellconx, erc_worstnwellcony;
INTBIG      erc_worstnwelledgex, erc_worstnwelledgey;

/* well areas */
#define NOWELLAREA ((WELLAREA *)-1)

typedef struct Iwellarea
{
	INTBIG            lx, hx, ly, hy;
	INTBIG            netnum;
	POLYGON           *poly;
	INTSML            layer;
	TECHNOLOGY        *tech;
	NODEPROTO         *facet;
	struct Iwellarea  *nextwellarea;
} WELLAREA;

WELLAREA *erc_wellareafree = NOWELLAREA;
WELLAREA *erc_firstwellarea = NOWELLAREA;

/* miscellaneous */

/* the meaning of "ERC_options" */
#define PWELLCONOPTIONS   03
#define PWELLCONPERAREA    0
#define PWELLCONONCE       1
#define PWELLCONIGNORE     2
#define NWELLCONOPTIONS  014
#define NWELLCONPERAREA    0
#define NWELLCONONCE      04
#define NWELLCONIGNORE   010

AIDENTRY   *erc_aid;						/* this tool */
NODEPROTO  *erc_facet;
INTBIG      erc_globalnetnumber;			/* global network number */
INTBIG      erc_attachednetnumber;			/* attached network number (above this floats) */
INTBIG      erc_optionskey;					/* key for "ERC_options" */
INTBIG      erc_options;					/* value of "ERC_options" */
INTBIG      erc_jobsize;					/* number of polygons to analyze */
INTBIG      erc_polyscrunched;				/* number of polygons analyzed */
INTBIG      erc_realrun;					/* nonzero if this is the real analysis */

/* prototypes for local routines */
WELLAREA    *erc_allocwellarea(void);
WELLCON     *erc_allocwellcon(void);
void         erc_analyzefacet(NODEPROTO *facet);
void         erc_freeallwellareas(void);
void         erc_freeallwellcons(void);
void         erc_freewellarea(WELLAREA *wa);
void         erc_freewellcon(WELLCON *wc);
void         erc_gatherwells(NODEPROTO *np, XARRAY trans);
void         erc_getpolygon(INTSML layer, TECHNOLOGY *tech, INTBIG *x, INTBIG *y, INTSML count);
void         erc_optionsdlog(void);
INTSML       erc_welllayer(TECHNOLOGY *tech, INTSML layer);

/************************ CONTROL ***********************/

/*
 * tool initialization
 */
void erc_init(INTBIG *argc, char *argv[], AIDENTRY *thisaid)
{
	/* only initialize during pass 1 */
	if (thisaid == NOAID || thisaid == 0) return;

	/* copy aid pointer during pass 1 */
	erc_aid = thisaid;
	erc_optionskey = makekey("ERC_options");
	DiaDeclareHook("ercopt", &erc_tablep, erc_optionsdlog);
}

void erc_done(void)
{
	erc_freeallwellcons();
	erc_freeallwellareas();
}

/*
 * Handle commands to the tool from the user.
 */
INTSML erc_set(INTSML count, char *par[])
{
	REGISTER INTSML l;
	REGISTER NODEPROTO *np;
	REGISTER char *pp;

	if (count == 0)
	{
		ttyputusage("tellaid erc analyze");
		return(0);
	}

	l = strlen(pp = par[0]);

	if (namesamen(pp, "analyze", l) == 0)
	{
		/* analyze facet */
		np = getcurfacet();
		if (np == NONODEPROTO) return(0);
		erc_analyzefacet(np);
		return(0);
	}

	ttyputbadusage("tellaid erc");
	return(1);
}

/******************** ANALYSIS ********************/

/*
 * Routine to analyze facet "facet" and build a list of ERC errors.
 */
void erc_analyzefacet(NODEPROTO *facet)
{
	REGISTER INTBIG errors, minsep, dist, layertype;
	REGISTER NODEPROTO *np;
	REGISTER TECHNOLOGY *tech;
	REGISTER char *abortseq, *errormsg;
	char errmsg[200];
	REGISTER WELLCON *wc, *owc;
	REGISTER VARIABLE *var;
	REGISTER PORTPROTO *pp, *opp;
	REGISTER WELLAREA *wa, *owa;
	REGISTER void *err;
	extern DIALOG us_progressdialog;

	/* make sure network tool is on */
	if ((net_aid->aidstate&AIDON) == 0)
	{
		ttyputerr(_("Network tool must be running...turning it on"));
		aidturnon(net_aid, 0);
		return;
	}

	/* announce start of analysis */
	(void)initinfstr();
	(void)addstringtoinfstr(_("Checking electrical rules..."));
	abortseq = getinterruptkey();
	if (abortseq != 0)
		(void)formatinfstr(_(" (type %s to abort)"), abortseq);
	ttyputmsg(returninfstr());
	if (DiaInitDialog(&us_progressdialog) != 0) return;
	DiaPercent(1, 0);
	DiaSetText(2, _("Preparing to analyze..."));

	/* initialize to analyze facet */
	initerrorlogging(_("ERC"));
	erc_freeallwellcons();
	erc_freeallwellareas();
	erc_worstpwelldist = erc_worstnwelldist = 0;
	erc_facet = facet;
	var = getvalkey((INTBIG)erc_aid, VAID, VINTEGER, erc_optionskey);
	if (var == NOVARIABLE) erc_options = 0; else
		erc_options = var->addr;

	/* put global network information on the ports of this facet */
	erc_globalnetnumber = 1;
	erc_attachednetnumber = 0;
	for(pp = facet->firstportproto; pp != NOPORTPROTO; pp = pp->nextportproto)
		pp->temp1 = 0;
	for(pp = facet->firstportproto; pp != NOPORTPROTO; pp = pp->nextportproto)
	{
		if (pp->temp1 != 0) continue;
		pp->temp1 = erc_globalnetnumber++;
		for(opp = pp->nextportproto; opp != NOPORTPROTO; opp = opp->nextportproto)
		{
			if (opp->temp1 != 0) continue;
			if (opp->network == pp->network) opp->temp1 = pp->temp1;
		}
	}

	/* reset count of primitives */
	for(tech = el_technologies; tech != NOTECHNOLOGY; tech = tech->nexttechnology)
		for(np = tech->firstnodeproto; np != NONODEPROTO; np = np->nextnodeproto)
			np->temp1 = 0;

	/* run through the work and count the number of polygons */
	erc_polyscrunched = 0;
	erc_realrun = 0;
	erc_gatherwells(facet, el_matid);
	erc_jobsize = erc_polyscrunched;

	/* now gather all of the geometry into the polygon merging system */
	DiaSetText(2, _("Gathering geometry..."));
	mergeinit();
	erc_realrun = 1;
	erc_polyscrunched = 0;
	erc_gatherwells(facet, el_matid);

	/* determine the number of polygon points in the analysis */
	DiaSetText(2, _("Analyzing..."));
	erc_polyscrunched = 0;
	erc_realrun = 0;
	mergedone(erc_getpolygon);
	erc_jobsize = erc_polyscrunched;

	/* do the analysis */
	erc_realrun = 1;
	erc_polyscrunched = 0;
	mergedone(erc_getpolygon);
	DiaPercent(1, 100);

	/* make sure all of the contacts are on the same net */
	for(wc = erc_firstwellcon; wc != NOWELLCON; wc = wc->nextwellcon)
	{
		if (wc->netnum == 0)
		{
			if (wc->fun == NPWELL) errormsg = _("P-Well contact is floating"); else
				errormsg = _("N-Well contact is floating");
			err = logerror(errormsg, erc_facet, 0);
			(void)addpointtoerror(err, wc->x, wc->y);
			continue;
		}
		for(owc = wc->nextwellcon; owc != NOWELLCON; owc = owc->nextwellcon)
		{
			if (owc->netnum == 0) continue;
			if (owc->fun != wc->fun) continue;
			if (owc->netnum == wc->netnum) continue;
			if (wc->fun == NPWELL) errormsg = _("P-Well contacts are not connected"); else
				errormsg = _("N-Well contacts are not connected");
			err = logerror(errormsg, erc_facet, 0);
			(void)addpointtoerror(err, wc->x, wc->y);
			(void)addpointtoerror(err, owc->x, owc->y);
			break;
		}
	}

	/* if just 1 N-Well contact is needed, see if it is there */
	if ((erc_options&NWELLCONOPTIONS) == NWELLCONONCE)
	{
		for(wc = erc_firstwellcon; wc != NOWELLCON; wc = wc->nextwellcon)
			if (wc->fun == NPSUBSTRATE) break;
		if (wc == NOWELLCON)
		{
			(void)logerror(_("No N-Well contact found in this facet"), erc_facet, 0);
		}
	}

	/* if just 1 P-Well contact is needed, see if it is there */
	if ((erc_options&PWELLCONOPTIONS) == PWELLCONONCE)
	{
		for(wc = erc_firstwellcon; wc != NOWELLCON; wc = wc->nextwellcon)
			if (wc->fun == NPWELL) break;
		if (wc == NOWELLCON)
		{
			(void)logerror(_("No P-well contact found in this facet"), erc_facet, 0);
		}
	}

	/* make sure the wells are separated properly */
	for(wa = erc_firstwellarea; wa != NOWELLAREA; wa = wa->nextwellarea)
	{
		for(owa = wa->nextwellarea; owa != NOWELLAREA; owa = owa->nextwellarea)
		{
			if (wa->tech != owa->tech || wa->layer != owa->layer) continue;
			if (wa->netnum == owa->netnum && wa->netnum >= 0)
				minsep = drcmindistance(wa->tech, facet->cell->lib, wa->layer, 0, wa->layer, 0, 1, 0, 0); else
					minsep = drcmindistance(wa->tech, facet->cell->lib, wa->layer, 0, wa->layer, 0, 0, 0, 0);
			if (minsep < 0) continue;
			dist = polyseparation(wa->poly, owa->poly);
			if (dist < minsep)
			{
				layertype = erc_welllayer(wa->tech, wa->layer);
				if (layertype == 0) continue;
				switch (layertype)
				{
					case 1:
						sprintf(errmsg, _("P-Well areas too close (are %s, should be %s)"),
							latoa(dist), latoa(minsep));
						break;
					case 2:
						sprintf(errmsg, _("N-Well areas too close (are %s, should be %s)"),
							latoa(dist), latoa(minsep));
						break;
					case 3:
						sprintf(errmsg, _("P-Select areas too close (are %s, should be %s)"),
							latoa(dist), latoa(minsep));
						break;
					case 4:
						sprintf(errmsg, _("N-Select areas too close (are %s, should be %s)"),
							latoa(dist), latoa(minsep));
						break;
				}
				err = logerror(errmsg, erc_facet, 0);
				(void)addpolytoerror(err, wa->poly);
				(void)addpolytoerror(err, owa->poly);
			}
		}
	}

#if 0			/* debugging to show all well areas */
	for(wa = erc_firstwellarea; wa != NOWELLAREA; wa = wa->nextwellarea)
	{
		layertype = erc_welllayer(wa->tech, wa->layer);
		if (layertype == 0) continue;
		switch (layertype)
		{
			case 1:
				err = logerror("P-Well area", erc_facet, 0);
				break;
			case 2:
				err = logerror("N-Well area", erc_facet, 0);
				break;
			case 3:
				err = logerror("P-Select area", erc_facet, 0);
				break;
			case 4:
				err = logerror("N-Select area", erc_facet, 0);
		}
		(void)addpolytoerror(err, wa->poly);
	}
#endif

	/* show the farthest distance from a well contact */
	if (erc_worstpwelldist > 0 || erc_worstnwelldist > 0)
	{
		(void)askaid(us_aid, "clear");
		if (erc_worstpwelldist > 0)
		{
			(void)askaid(us_aid, "show-line", erc_worstpwellconx, erc_worstpwellcony,
				erc_worstpwelledgex, erc_worstpwelledgey, erc_facet);
			ttyputmsg(_("Farthest distance from a P-Well contact is %s"),
				latoa(erc_worstpwelldist));
		}
		if (erc_worstnwelldist > 0)
		{
			(void)askaid(us_aid, "show-line", erc_worstnwellconx, erc_worstnwellcony,
				erc_worstnwelledgex, erc_worstnwelledgey, erc_facet);
			ttyputmsg(_("Farthest distance from a N-Well contact is %s"),
				latoa(erc_worstnwelldist));
		}
	}

	/* report the number of transistors */
	for(tech = el_technologies; tech != NOTECHNOLOGY; tech = tech->nexttechnology)
	{
		for(np = tech->firstnodeproto; np != NONODEPROTO; np = np->nextnodeproto)
		{
			if (np->temp1 == 0) continue;
			switch ((np->userbits&NFUNCTION) >> NFUNCTIONSH)
			{
				case NPTRANMOS:   case NPTRA4NMOS:
				case NPTRADMOS:   case NPTRA4DMOS:
				case NPTRAPMOS:   case NPTRA4PMOS:
				case NPTRANPN:    case NPTRA4NPN:
				case NPTRAPNP:    case NPTRA4PNP:
				case NPTRANJFET:  case NPTRA4NJFET:
				case NPTRAPJFET:  case NPTRA4PJFET:
				case NPTRADMES:   case NPTRA4DMES:
				case NPTRAEMES:   case NPTRA4EMES:
				case NPTRANSREF:
				case NPTRANS:
					ttyputmsg(_("%ld nodes of type %s"), np->temp1, describenodeproto(np));
					break;
			}
		}
	}

	/* report the number of networks */
	DiaDoneDialog();
	ttyputmsg(_("%ld networks found"), erc_globalnetnumber);
	
	/* report the number of errors found */
	errors = numerrors();
	if (errors == 0)
		ttyputmsg(_("No ERC errors found")); else
			ttyputmsg(_("FOUND %ld ERC %s"), errors, makeplural(_("ERROR"), errors));
	termerrorlogging();
}

/*
 * Recursive helper routine to gather all well areas in facet "np"
 * (transformed by "trans").
 */
void erc_gatherwells(NODEPROTO *np, XARRAY trans)
{
	REGISTER NODEINST *ni;
	REGISTER NODEPROTO *subnp;
	REGISTER ARCINST *ai;
	REGISTER PORTARCINST *pi;
	REGISTER PORTEXPINST *pe;
	REGISTER PORTPROTO *pp, *opp;
	REGISTER NETWORK *net;
	REGISTER INTBIG i, tot, nfun;
	static INTBIG checkstop = 0;
	XARRAY localrot, localtran, final, subrot;
	static POLYGON *poly = NOPOLYGON;
	REGISTER WELLCON *wc;

	if (poly == NOPOLYGON) poly = allocstaticpolygon(4, erc_aid->cluster);

	/* grab global network information from the ports */
	if (erc_realrun != 0)
	{
		for(net = np->firstnetwork; net != NONETWORK; net = net->nextnetwork)
			net->temp1 = 0;
		for(pp = np->firstportproto; pp != NOPORTPROTO; pp = pp->nextportproto)
			pp->network->temp1 = pp->temp1;
		for(net = np->firstnetwork; net != NONETWORK; net = net->nextnetwork)
		{
			if (net->temp1 == 0) net->temp1 = erc_globalnetnumber++;
		}
		if (erc_attachednetnumber == 0) erc_attachednetnumber = erc_globalnetnumber;
		DiaPercent(1, erc_polyscrunched * 100 / erc_jobsize);
	}

	for(ni = np->firstnodeinst; ni != NONODEINST; ni = ni->nextnodeinst)
	{
		subnp = ni->proto;
		if (erc_realrun != 0)
		{
			makerot(ni, localrot);
			transmult(localrot, trans, final);
		}
		if (subnp->primindex == 0)
		{
			if (erc_realrun != 0)
			{
				if (stopping(STOPREASONERC)) return;

				/* propagate global network numbers into this instance */
				for(pp = subnp->firstportproto; pp != NOPORTPROTO; pp = pp->nextportproto)
					pp->temp1 = 0;
				for(pi = ni->firstportarcinst; pi != NOPORTARCINST; pi = pi->nextportarcinst)
					pi->proto->temp1 = pi->conarcinst->network->temp1;
				for(pe = ni->firstportexpinst; pe != NOPORTEXPINST; pe = pe->nextportexpinst)
					pe->proto->temp1 = pe->exportproto->network->temp1;
				for(pp = subnp->firstportproto; pp != NOPORTPROTO; pp = pp->nextportproto)
				{
					if (pp->temp1 != 0) continue;
					for(opp = subnp->firstportproto; opp != NOPORTPROTO; opp = opp->nextportproto)
						if (opp != pp && opp->temp1 != 0 && opp->network == pp->network) break;
					if (opp != NOPORTPROTO) pp->temp1 = opp->temp1; else
						pp->temp1 = erc_globalnetnumber++;
				}

				maketrans(ni, localtran);
				transmult(localtran, final, subrot);
			}
			erc_gatherwells(subnp, subrot);
		} else
		{
			if (erc_realrun != 0)
			{
				checkstop++;
				if ((checkstop%10) == 0)
				{
					if (stopping(STOPREASONERC)) return;
				}
				nfun = nodefunction(ni, 0);
				ni->proto->temp1++;

				if (nfun == NPWELL || nfun == NPSUBSTRATE)
				{
					wc = erc_allocwellcon();
					if (wc == NOWELLCON) return;
					xform((ni->lowx + ni->highx) / 2, (ni->lowy + ni->highy) / 2,
						&wc->x, &wc->y, final);
					wc->np = subnp;
					wc->fun = nfun;
					if (ni->firstportarcinst != NOPORTARCINST)
					{
						wc->netnum = ni->firstportarcinst->conarcinst->network->temp1;
					} else if (ni->firstportexpinst != NOPORTEXPINST)
					{
						wc->netnum = ni->firstportexpinst->exportproto->network->temp1;
					} else wc->netnum = 0;
					wc->nextwellcon = erc_firstwellcon;
					erc_firstwellcon = wc;
				}
			}

			tot = nodepolys(ni, 0, NOWINDOWPART);
			erc_polyscrunched += tot;
			if (erc_realrun != 0)
			{
				for(i=0; i<tot; i++)
				{
					shapenodepoly(ni, i, poly);
					if (erc_welllayer(poly->tech, poly->layer) == 0) continue;
					xformpoly(poly, final);
					mergestorepolygon(poly->layer, poly->tech, poly);
				}
			}
		}
	}

	for(ai = np->firstarcinst; ai != NOARCINST; ai = ai->nextarcinst)
	{
		checkstop++;
		if ((checkstop%10) == 0)
		{
			if (stopping(STOPREASONERC)) return;
		}
		tot = arcpolys(ai, NOWINDOWPART);
		erc_polyscrunched += tot;
		if (erc_realrun != 0)
		{
			for(i=0; i<tot; i++)
			{
				shapearcpoly(ai, i, poly);
				if (erc_welllayer(poly->tech, poly->layer) == 0) continue;
				xformpoly(poly, trans);
				mergestorepolygon(poly->layer, poly->tech, poly);
			}
		}
	}
}

/*
 * Coroutine of the polygon merging package that is given a merged polygon with
 * "count" points in (x,y) with area "area".
 */
void erc_getpolygon(INTSML layer, TECHNOLOGY *tech, INTBIG *x, INTBIG *y, INTSML count)
{
	REGISTER INTSML i, first, welltype;
	REGISTER INTBIG dist, bestdist, desiredcontact, perareabit, prev, xp, yp;
	REGISTER WELLCON *wc, *bestwc;
	REGISTER WELLAREA *wa;
	REGISTER void *err;
	REGISTER char *nocontacterror;
	static INTBIG checkstop = 0;

	erc_polyscrunched += count;
	if (erc_realrun == 0) return;

	checkstop++;
	if ((checkstop%10) == 0)
	{
		if (stopping(STOPREASONERC)) return;
	} else
	{
		if (el_pleasestop != 0) return;
	}
	DiaPercent(1, erc_polyscrunched * 100 / erc_jobsize);

	/* save the polygon */
	wa = erc_allocwellarea();
	wa->poly = allocpolygon(count, erc_aid->cluster);
	for(i=0; i<count; i++)
	{
		wa->poly->xv[i] = x[i];
		wa->poly->yv[i] = y[i];
	}
	wa->poly->count = count;
	wa->poly->style = FILLED;
	wa->layer = layer;
	wa->netnum = 0;
	wa->tech = tech;
	getbbox(wa->poly, &wa->lx, &wa->hx, &wa->ly, &wa->hy);
	wa->nextwellarea = erc_firstwellarea;
	erc_firstwellarea = wa;

	/* special checks for the well areas */
	welltype = erc_welllayer(tech, layer);
	if (welltype != 1 && welltype != 2) return;
	if (welltype == 1)
	{
		/* P-well */
		desiredcontact = NPWELL;
		perareabit = PWELLCONPERAREA;
		nocontacterror = _("No P-Well contact found in this area");
	} else
	{
		/* N-Well */
		desiredcontact = NPSUBSTRATE;
		perareabit = NWELLCONPERAREA;
		nocontacterror = _("No N-Well contact found in this area");
	}

	/* find a contact in the area */
	for(wc = erc_firstwellcon; wc != NOWELLCON; wc = wc->nextwellcon)
	{
		if (wc->fun != desiredcontact) continue;
		if (isinside(wc->x, wc->y, wa->poly) == 0) continue;
		wa->netnum = wc->netnum;
		break;
	}

	/* if no contact, issue appropriate errors */
	if (wc == NOWELLCON)
	{
		if ((erc_options&PWELLCONOPTIONS) == perareabit)
		{
			err = logerror(nocontacterror, erc_facet, 0);
			(void)addpolytoerror(err, wa->poly);
		}
		return;
	}

	/* find the worst distance to the edge of the area */
	for(i=0; i<count*2; i++)
	{
		/* figure out which point is being analyzed */
		if (i < count)
		{
			if (i == 0) prev = count-1; else prev = i-1;
			xp = (x[prev] + x[i]) / 2;
			yp = (y[prev] + y[i]) / 2;
		} else
		{
			xp = x[i-count];
			yp = y[i-count];
		}

		/* find the closest contact to this point */
		first = 1;
		bestdist = 0;
		for(wc = erc_firstwellcon; wc != NOWELLCON; wc = wc->nextwellcon)
		{
			if (wc->fun != desiredcontact) continue;
			if (isinside(wc->x, wc->y, wa->poly) == 0) continue;
			dist = computedistance(wc->x, wc->y, xp, yp);
			if (first != 0 || dist < bestdist)
			{
				bestdist = dist;
				bestwc = wc;
			}
			first = 0;
		}
		if (first != 0) continue;

		/* accumulate worst distances to edges */
		if (welltype == 1)
		{
			if (bestdist > erc_worstpwelldist)
			{
				erc_worstpwelldist = bestdist;
				erc_worstpwellconx = bestwc->x;
				erc_worstpwellcony = bestwc->y;
				erc_worstpwelledgex = xp;
				erc_worstpwelledgey = yp;
			}
		} else
		{
			if (bestdist > erc_worstnwelldist)
			{
				erc_worstnwelldist = bestdist;
				erc_worstnwellconx = bestwc->x;
				erc_worstnwellcony = bestwc->y;
				erc_worstnwelledgex = xp;
				erc_worstnwelledgey = yp;
			}
		}
	}
}

/*
 * Routine to return nonzero if layer "layer" in technology "tech" is a well/select layer.
 * Returns:
 *   1: P-Well
 *   2: N-Well
 *   3: P-Select
 *   4: N-Select
 */
INTSML erc_welllayer(TECHNOLOGY *tech, INTSML layer)
{
	REGISTER INTBIG fun, funpure;

	fun = layerfunction(tech, layer);
	funpure = fun & LFTYPE;
	if (funpure == LFWELL)
	{
		if ((fun&LFPTYPE) != 0) return(1);
		return(2);
	}
	if (funpure == LFSUBSTRATE || funpure == LFIMPLANT)
	{
		if ((fun&LFPTYPE) != 0) return(3);
		return(4);
	}
	return(0);
}

void erc_freeallwellcons(void)
{
	REGISTER WELLCON *wc;

	while (erc_firstwellcon != NOWELLCON)
	{
		wc = erc_firstwellcon;
		erc_firstwellcon = wc->nextwellcon;
		efree((char *)wc);
	}
}

WELLCON *erc_allocwellcon(void)
{
	REGISTER WELLCON *wc;

	if (erc_wellconfree != NOWELLCON)
	{
		wc = erc_wellconfree;
		erc_wellconfree = wc->nextwellcon;
	} else
	{
		wc = (WELLCON *)emalloc(sizeof (WELLCON), erc_aid->cluster);
		if (wc == 0) return(NOWELLCON);
	}
	return(wc);
}

void erc_freewellcon(WELLCON *wc)
{
	wc->nextwellcon = erc_wellconfree;
	erc_wellconfree = wc;
}

void erc_freeallwellareas(void)
{
	REGISTER WELLAREA *wa;

	while (erc_firstwellarea != NOWELLAREA)
	{
		wa = erc_firstwellarea;
		erc_firstwellarea = wa->nextwellarea;
		freepolygon(wa->poly);
		efree((char *)wa);
	}
}

WELLAREA *erc_allocwellarea(void)
{
	REGISTER WELLAREA *wa;

	if (erc_wellareafree != NOWELLAREA)
	{
		wa = erc_wellareafree;
		erc_wellareafree = wa->nextwellarea;
	} else
	{
		wa = (WELLAREA *)emalloc(sizeof (WELLAREA), erc_aid->cluster);
		if (wa == 0) return(NOWELLAREA);
	}
	return(wa);
}

void erc_freewellarea(WELLAREA *wa)
{
	wa->nextwellarea = erc_wellareafree;
	erc_wellareafree = wa;
}

/******************** OPTIONS DIALOG ********************/

/* ERC: Options Dialog */
DIALOGITEM erc_optionsdialogitems[] =
{
 /*  1 */ {0, {108,396,132,460}, BUTTON, N_("OK")},
 /*  2 */ {0, {108,296,132,360}, BUTTON, N_("Cancel")},
 /*  3 */ {0, {32,8,48,228}, RADIO, N_("Must have contact in every area")},
 /*  4 */ {0, {56,8,72,228}, RADIO, N_("Must have at least 1 contact")},
 /*  5 */ {0, {80,8,96,228}, RADIO, N_("Do not check for contacts")},
 /*  6 */ {0, {8,44,24,144}, MESSAGE, N_("For P-Well:")},
 /*  7 */ {0, {32,248,48,480}, RADIO, N_("Must have contact in every area")},
 /*  8 */ {0, {56,248,72,480}, RADIO, N_("Must have at least 1 contact")},
 /*  9 */ {0, {80,248,96,480}, RADIO, N_("Do not check for contacts")},
 /* 10 */ {0, {8,284,24,384}, MESSAGE, N_("For N-Well:")}
};
DIALOG erc_optionsdialog = {{50,75,191,564}, N_("ERC Options"), 0, 10, erc_optionsdialogitems};

/* special items for the "ERC Options" dialog: */
#define DERO_PWELLEACH   3		/* PWell contact in each area (radio) */
#define DERO_PWELLONCE   4		/* PWell contact once (radio) */
#define DERO_PWELLNOT    5		/* PWell contact not needed (radio) */
#define DERO_NWELLEACH   7		/* NWell contact in each area (radio) */
#define DERO_NWELLONCE   8		/* NWell contact once (radio) */
#define DERO_NWELLNOT    9		/* NWell contact not needed (radio) */

void erc_optionsdlog(void)
{
	REGISTER INTBIG itemHit, ercoptions, oldoptions;
	REGISTER VARIABLE *var;

	if (DiaInitDialog(&erc_optionsdialog) != 0) return;
	var = getvalkey((INTBIG)erc_aid, VAID, VINTEGER, erc_optionskey);
	if (var == NOVARIABLE) ercoptions = 0; else
		ercoptions = var->addr;
	oldoptions = ercoptions;
	switch (ercoptions & PWELLCONOPTIONS)
	{
		case PWELLCONPERAREA: DiaSetControl(DERO_PWELLEACH, 1);   break;
		case PWELLCONONCE:    DiaSetControl(DERO_PWELLONCE, 1);   break;
		case PWELLCONIGNORE:  DiaSetControl(DERO_PWELLNOT, 1);    break;
	}
	switch (ercoptions & NWELLCONOPTIONS)
	{
		case NWELLCONPERAREA: DiaSetControl(DERO_NWELLEACH, 1);   break;
		case NWELLCONONCE:    DiaSetControl(DERO_NWELLONCE, 1);   break;
		case NWELLCONIGNORE:  DiaSetControl(DERO_NWELLNOT, 1);    break;
	}

	/* loop until done */
	for(;;)
	{
		itemHit = DiaNextHit();
		if (itemHit == OK || itemHit == CANCEL) break;
		if (itemHit == DERO_PWELLEACH || itemHit == DERO_PWELLONCE ||
			itemHit == DERO_PWELLNOT)
		{
			DiaSetControl(DERO_PWELLEACH, 0);
			DiaSetControl(DERO_PWELLONCE, 0);
			DiaSetControl(DERO_PWELLNOT, 0);
			DiaSetControl(itemHit, 1);
			continue;
		}
		if (itemHit == DERO_NWELLEACH || itemHit == DERO_NWELLONCE ||
			itemHit == DERO_NWELLNOT)
		{
			DiaSetControl(DERO_NWELLEACH, 0);
			DiaSetControl(DERO_NWELLONCE, 0);
			DiaSetControl(DERO_NWELLNOT, 0);
			DiaSetControl(itemHit, 1);
			continue;
		}
	}

	if (itemHit != CANCEL)
	{
		/* change state of ERC */
		ercoptions = 0;
		if (DiaGetControl(DERO_PWELLEACH) != 0) ercoptions |= PWELLCONPERAREA; else
			if (DiaGetControl(DERO_PWELLONCE) != 0) ercoptions |= PWELLCONONCE; else
				ercoptions |= PWELLCONIGNORE;
		if (DiaGetControl(DERO_NWELLEACH) != 0) ercoptions |= NWELLCONPERAREA; else
			if (DiaGetControl(DERO_NWELLONCE) != 0) ercoptions |= NWELLCONONCE; else
				ercoptions |= NWELLCONIGNORE;
		if (ercoptions != oldoptions)
			(void)setvalkey((INTBIG)erc_aid, VAID, erc_optionskey, ercoptions, VINTEGER);
	}
	DiaDoneDialog();
}

#endif  /* ERCAID - at top */
