/*
 * Electric(tm) VLSI Design System
 *
 * File: simwindow.c
 * Simulation aid: signal window control
 * Written by: Steven M. Rubin, Static Free Software
 *
 * Copyright (c) 2000 Static Free Software.
 *
 * Electric(tm) is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Electric(tm) is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Electric(tm); see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, Mass 02111-1307, USA.
 *
 * Static Free Software
 * 4119 Alpine Road
 * Portola Valley, California 94028
 * info@staticfreesoft.com
 */

#include "config.h"
#if SIMAID

#include "global.h"
#include "egraphics.h"
#include "sim.h"
#include "simals.h"
#include "tecgen.h"
#include "tecschem.h"
#include "tecart.h"
#include "usr.h"
#include "usrtrack.h"

/*
 *     WINDOW CONTROL:
 * sim_window_init()                            Initialize simulation window system (do only once)
 * sim_window_term()                            Terminate simulation window system (do only once)
 * INTSML sim_window_create(l, np, chw, chs)    Make simulation window with "l" lines
 *                                              in facet "np" charhandlers "chw/s" (l=0 to restore)
 * void sim_window_stopsimulation()             Stops simulation
 * INTBIG sim_window_isactive(np)               Returns nonzero if simulating, returns facet if so
 * sim_window_redraw()                          Redraws simulation window
 * sim_window_setlines(count)                   Sets number of lines in simulation window
 * INTBIG sim_window_getlines()                 Returns number of lines in sim window
 * sim_window_setvislines(count)                Sets number of visible lines in simulation window
 * INTBIG sim_window_gettopline()               Returns top line in simulation window
 * sim_window_settopline(count)                 Sets top visible line in simulation window
 * INTBIG sim_window_getvislines()              Returns number of visible lines in sim window
 * sim_window_savegraph()                       Preserves plot in database
 * sim_window_displaycolor(s, c)                Set color of strength/state "s" to "c"
 *
 *     CONTROL OF TRACES:
 * INTBIG sim_window_newtrace(line, name, data) Creates trace at "line" called "name" with user "data"
 * INTSML sim_window_buscommand()               Creates a bus from currently selected signals
 * sim_window_loaddigtrace(tr, num, time, sta)  Loads trace "tr" with "num" digital signals "time/sta"
 * sim_window_loadanatrace(tr, num, time, val)  Loads trace "tr" with "num" analog signals "time/val"
 * sim_window_setanarange(low, high)            Sets analog display range from "low" to "high"
 * sim_window_getanarange(low, high)            Gets analog display range
 * sim_window_setanaextents(low, high)          Sets fixed analog extent from "low" to "high"
 * sim_window_getanaextents(low, high)          Gets fixed analog extent
 * sim_window_settraceline(tr, line)            Sets line number for trace "tr"
 * sim_window_killtrace(tr)                     Deletes trace "tr"
 * sim_window_killalltraces()                   Deletes all traces
 *
 *     TRACE INFORMATION:
 * INTBIG *sim_window_getbustraces(tr)          Returns 0-terminated list of traces in bus "tr"
 * INTBIG sim_window_gettraceline(tr)           Returns line number for trace "tr"
 * char *sim_window_gettracename(tr)            Returns name for trace "tr"
 * INTBIG sim_window_gettracedata(tr)           Returns user data for trace "tr"
 * float sim_window_getanatracevalue(tr, time)  Returns analog value for trace "tr" at time "time"
 * INTBIG sim_window_findtrace(name)            Locates the trace with this name
 * sim_window_inittraceloop()                   Begin search of all traces
 * sim_window_inittraceloop2()                  Begin search of all traces
 * INTBIG sim_window_nexttraceloop()            Returns next trace (0 when done)
 * INTBIG sim_window_nexttraceloop2()           Returns next trace (0 when done)
 *
 *     TRACE HIGHLIGHTING:
 * sim_window_cleartracehighlight()             Unhighlights all traces
 * sim_window_addhighlighttrace(tr)             Adds trace "tr" to highlighting
 * sim_window_deletehighlighttrace(tr)          Removes trace "tr" from highlighting
 * INTBIG sim_window_gethighlighttrace()        Returns highlighted trace (0 if none)
 * INTBIG *sim_window_gethighlighttraces()      Returns 0-terminated list of highlighted traces
 *
 *     TIME AND CURSOR CONTROL:
 * sim_window_setmaincursor(time)               Sets position of main cursor
 * float sim_window_getmaincursor()             Returns position of main cursor
 * sim_window_setextensioncursor(time)          Sets position of extension cursor
 * float sim_window_getextensioncursor()        Returns position of extension cursor
 * sim_window_settimerange(min, max)            Sets window time range from "min" to "max"
 * sim_window_gettimerange(min, max)            Returns window time range
 * sim_window_gettimeextents(min, max)          Returns data time range
 */

#define NODISCHANNEL ((DISCHANNEL *)-1)

/* the meaning of DISCHANNEL->flags */
#define TRACEWANTSELECT  01		/* set if this trace wants to be selected */
#define TRACETYPE        06		/* the type of this trace */
#define TRACEISDIGITAL    0		/*     this trace is a digital signal */
#define TRACEISANALOG    02		/*     this trace is an analog signal */
#define TRACEISBUS       04		/*     this trace is a bus */
#define TRACEISOPENBUS   06		/*     this trace is an opened bus */
#define TRACENOTDRAWN   010		/* set if trace is not to be drawn */

typedef struct Idischannel
{
	char     *name;			/* name of this trace */
	char     *origname;		/* original name of this trace */
	INTBIG    flags;		/* state bits for this trace (see above) */
	INTBIG    nodeptr;		/* "user" data for this trace */
	INTBIG    busindex;		/* index of this trace in the bus (when part of a bus) */
	INTBIG    color;		/* color of this trace (analog only) */
	INTBIG    nameoff;		/* offset of trace name (when many per line) */
	INTBIG    position;		/* line number of this trace in window */
	INTBIG    numsteps;		/* number of steps along this trace */
	INTBIG    timelimit;	/* maximum times allocated for this trace */
	INTBIG    statelimit;	/* maximum states allocated for this trace */
	INTBIG    valuelimit;	/* maximum values allocated for this trace */
	float    *timearray;	/* time steps along this trace */
	INTBIG   *statearray;	/* states along this trace (digital) */
	float    *valuearray;	/* values along this trace (analog) */
	struct Idischannel *buschannel;
	struct Idischannel *nextdischannel;
} DISCHANNEL;

static GRAPHICS sim_window_desc = {LAYERA, ALLOFF, SOLIDC, SOLIDC,
	{0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF}, NOVARIABLE, 0};
static INTBIG sim_window_anacolor[] = {RED, GREEN, BLUE, CYAN, MAGENTA, YELLOW, GRAY, ORANGE,
	PURPLE, BROWN, LGRAY, DGRAY, LRED, DRED, LGREEN, DGREEN, LBLUE, DBLUE, 0};

/*
 * In X:
 *    0 to 112 is signal name
 *      113    is bar
 *   114-767 is signal waveform
 * In Y:
 *    0 to  30 is axis label
 *   31 to 570 is signal waveforms
 *  571 to 600 is cursor value
 */
static INTBIG        sim_window_curx, sim_window_cury;
static INTBIG        sim_window_textsize;
static INTBIG        sim_window_lines;
static INTBIG        sim_window_vislines;
static INTBIG        sim_window_topline;
static INTBIG        sim_window_txthei;
static INTBIG        sim_window_offx, sim_window_offy;
static INTBIG        sim_window_basex, sim_window_basey;
       INTBIG        sim_window_state;		/* cached key for "SIM_window_state" */
       INTBIG        sim_window_signal_order;	/* cached key for "SIM_window_signal_order" */
static float         sim_window_maincursor, sim_window_extensioncursor;
static float         sim_window_mintime, sim_window_maxtime;
static float         sim_window_analow, sim_window_anahigh, sim_window_anarange;
static float         sim_window_anadislow, sim_window_anadishigh, sim_window_anadisrange;
static INTBIG        sim_window_highlightedtracecount;
static INTBIG        sim_window_highlightedtracetotal = 0;
static DISCHANNEL  **sim_window_highlightedtraces;
static INTBIG        sim_window_bustracetotal = 0;
static DISCHANNEL  **sim_window_bustraces;
static DISCHANNEL   *sim_window_plot;
static DISCHANNEL   *sim_window_plotfree;
static DISCHANNEL   *sim_window_loop;
static DISCHANNEL   *sim_window_loop2;
static INTBIG        sim_colorstrengthoff;		/* color of off-strength */
static INTBIG        sim_colorstrengthnode;		/* color of node-strength */
static INTBIG        sim_colorstrengthgate;		/* color of gate-strength */
static INTBIG        sim_colorstrengthpower;	/* color of power-strength */
static INTBIG        sim_colorlevellow;			/* color of low-levels */
static INTBIG        sim_colorlevelhigh;		/* color of high-levels */
static INTBIG        sim_colorlevelundef;		/* color of power-levels */
static WINDOWPART   *sim_waveformwindow;		/* current window when arrows are clicked */
static INTBIG        sim_waveformsliderpart;	/* 0:down arrow 1:below thumb 2:thumb 3:above thumb 4:up arrow */
static INTBIG        sim_window_tracedragyoff;	/* Y offset when dragging */
static INTBIG        sim_window_initialx;		/* initial X coordinate when dragging */
static INTBIG        sim_window_initialy;		/* initial Y coordinate when dragging */
static INTBIG        sim_window_currentdragy;	/* current Y coordinate when dragging */
static DISCHANNEL   *sim_window_tracedragtr;	/* signal being dragged */

/* prototypes for local routines */
void        sim_window_addsegment(WINDOWPART*, INTBIG*, INTBIG*, INTBIG, INTBIG, INTBIG, INTBIG);
void        sim_window_addtrace(DISCHANNEL *tr, DISCHANNEL *aftertr);
DISCHANNEL *sim_window_allocdischannel(void);
void        sim_window_buttonhandler(WINDOWPART*, INTSML, INTSML, INTSML);
void        sim_window_drawbox(WINDOWPART*, INTBIG, INTBIG, INTBIG, INTBIG);
void        sim_window_drawcstring(WINDOWPART*, char*);
void        sim_window_drawcursors(WINDOWPART*);
void        sim_window_drawgraph(WINDOWPART*);
void        sim_window_drawhcstring(WINDOWPART *w, char *s);
void        sim_window_drawrstring(WINDOWPART*, char*);
void        sim_window_drawtimescale(WINDOWPART*);
void        sim_window_drawto(WINDOWPART*, INTBIG, INTBIG);
void        sim_window_drawtracedrag(WINDOWPART*, INTSML on);
void        sim_window_drawtraceselect(WINDOWPART *wavewin, INTBIG y1, INTBIG y2, INTSML on);
void        sim_window_drawulstring(WINDOWPART *w, char *s);
INTSML      sim_window_eachbdown(INTBIG, INTBIG);
INTSML      sim_window_eachdown(INTBIG x, INTBIG y);
INTSML      sim_window_eachwdown(INTBIG, INTBIG);
WINDOWPART *sim_window_findschematics(void);
WINDOWPART *sim_window_findwaveform(void);
INTSML      sim_window_horizontalslider(INTBIG x, INTBIG y);
void        sim_window_mapcoord(WINDOWPART*, INTBIG*, INTBIG*);
void        sim_window_moveto(INTBIG, INTBIG);
char       *sim_window_namesignals(INTBIG count, DISCHANNEL **list);
void        sim_window_plottrace(DISCHANNEL *tr, WINDOWPART *wavewin, INTBIG highlight);
void        sim_window_redisphandler(WINDOWPART*);
void        sim_window_removetrace(DISCHANNEL *tr);
void        sim_window_savesignalorder(void);
void        sim_window_scaletowindow(WINDOWPART*, INTBIG*, INTBIG*);
INTSML      sim_window_seleachdown(INTBIG x, INTBIG y);
void        sim_window_setcolor(INTBIG);
void        sim_window_setmask(INTSML);
void        sim_window_setviewport(WINDOWPART*);
INTBIG      sim_window_timetoxpos(float);
INTSML      sim_window_verticalslider(INTBIG x, INTBIG y);
void        sim_window_writetracename(WINDOWPART*, DISCHANNEL*);
float       sim_window_xpostotime(INTBIG);

/********************************* WINDOW CONTROL *********************************/

/*
 * routine to initialize the simulation window
 */
void sim_window_init(void)
{
	sim_window_signal_order = makekey("SIM_window_signal_order");

	sim_window_mintime = 0.0;
	sim_window_maxtime = 0.0000005f;
	sim_window_maincursor = 0.0000002f;
	sim_window_extensioncursor = 0.0000003f;
	sim_window_lines = 0;
	sim_window_vislines = 0;
	sim_window_topline = 0;
	sim_window_plot = NODISCHANNEL;
	sim_window_plotfree = NODISCHANNEL;
	sim_window_highlightedtracecount = 0;
	(void)setvalkey((INTBIG)sim_aid, VAID, sim_window_state,
		FULLSTATE | SHOWWAVEFORM, VINTEGER|VDONTSAVE);
	sim_colorstrengthoff = BLUE;
	sim_colorstrengthnode = GREEN;
	sim_colorstrengthgate = MAGENTA;
	sim_colorstrengthpower = BLACK;
	sim_colorlevellow = BLUE;
	sim_colorlevelhigh = MAGENTA;
	sim_colorlevelundef = BLACK;
}

/*
 * routine to terminate the simulation window
 */
void sim_window_term(void)
{
	REGISTER DISCHANNEL *tr;

	sim_window_killalltraces();

	while (sim_window_plotfree != NODISCHANNEL)
	{
		tr = sim_window_plotfree;
		sim_window_plotfree = sim_window_plotfree->nextdischannel;
		if (tr->timelimit > 0) efree((char *)tr->timearray);
		if (tr->statelimit > 0) efree((char *)tr->statearray);
		if (tr->valuelimit > 0) efree((char *)tr->valuearray);
		efree((char *)tr);
	}
	if (sim_window_highlightedtracetotal > 0) efree((char *)sim_window_highlightedtraces);
	if (sim_window_bustracetotal > 0) efree((char *)sim_window_bustraces);
}

/*
 * routine to start a simulation window with room for "linecount" traces and associated facet "np".
 * The character handler to use in the waveform window is in "charhandlerwave", the character
 * handler to use in the schematic/layout window is in "charhandlerschem".  Returns nonzero on error.
 */
INTSML sim_window_create(INTSML linecount, NODEPROTO *np, INTSML (*charhandlerwave)(WINDOWPART*, INTSML),
	INTSML (*charhandlerschem)(WINDOWPART*, INTSML))
{
	REGISTER WINDOWPART *wavewin, *schemwin;
	REGISTER NODEPROTO *simnp;
	REGISTER INTBIG windowstate;
	REGISTER VARIABLE *var;

	/* sorry, can't have more than 1 simulation window at a time */
	wavewin = sim_window_findwaveform();
	schemwin = sim_window_findschematics();
	simnp = NONODEPROTO;
	if (wavewin != NOWINDOWPART && (wavewin->state&WINDOWSIMULATING) != 0)
		simnp = wavewin->curnodeproto;
	if (schemwin != NOWINDOWPART) simnp = schemwin->curnodeproto;
	if (simnp != NONODEPROTO && simnp != np)
	{
		ttyputerr("Sorry, already simulating facet %s", describenodeproto(simnp));
		return(1);
	}

	/* reinitialize waveform display if number of lines is specified */
	if (linecount > 0)
	{
		sim_window_lines = linecount;
		if (sim_window_lines < 1) sim_window_lines = 1;
		sim_window_killalltraces();
		sim_window_topline = 0;
	}

	/* determine the number of visible signals */
	sim_window_vislines = sim_window_lines;

	/* create waveform window if requested */
	if (charhandlerwave != 0)
	{
		if (wavewin == NOWINDOWPART)
		{
			/* get window state information */
			var = getvalkey((INTBIG)sim_aid, VAID, VINTEGER, sim_window_state);
			if (var == NOVARIABLE) windowstate = 0; else windowstate = var->addr;
			switch (windowstate&WAVEPLACE)
			{
				case WAVEPLACECAS:
					wavewin = (WINDOWPART *)askaid(us_aid, "window-new");
					break;
				case WAVEPLACETHOR:
					wavewin = (WINDOWPART *)askaid(us_aid, "window-horiz-new");
					break;
				case WAVEPLACETVER:
					wavewin = (WINDOWPART *)askaid(us_aid, "window-vert-new");
					break;
			}
			if (wavewin == NOWINDOWPART) return(1);
		}
		sim_window_setviewport(wavewin);

		startobjectchange((INTBIG)wavewin, VWINDOWPART);
		(void)setval((INTBIG)wavewin, VWINDOWPART, "state", (wavewin->state & ~WINDOWTYPE) | WAVEFORMWINDOW |
			WINDOWSIMULATING, VINTEGER);
		(void)setval((INTBIG)wavewin, VWINDOWPART, "curnodeproto", (INTBIG)np, VNODEPROTO);
		(void)setval((INTBIG)wavewin, VWINDOWPART, "buttonhandler", (INTBIG)sim_window_buttonhandler, VADDRESS);
		(void)setval((INTBIG)wavewin, VWINDOWPART, "charhandler", (INTBIG)charhandlerwave, VADDRESS);
		(void)setval((INTBIG)wavewin, VWINDOWPART, "redisphandler", (INTBIG)sim_window_redisphandler, VADDRESS);
		endobjectchange((INTBIG)wavewin, VWINDOWPART);
		us_setfacetname(wavewin);
	}

	/* find and setup associated schematics/layout window if requested */
	if (charhandlerschem != 0)
	{
		for(schemwin = el_topwindowpart; schemwin != NOWINDOWPART;
			schemwin = schemwin->nextwindowpart)
		{
			if ((schemwin->state&WINDOWTYPE) != DISPWINDOW &&
				(schemwin->state&WINDOWTYPE) != DISP3DWINDOW) continue;
			if (schemwin->curnodeproto == np) break;
		}
		if (schemwin != NOWINDOWPART)
		{
			startobjectchange((INTBIG)schemwin, VWINDOWPART);
			(void)setval((INTBIG)schemwin, VWINDOWPART, "state", schemwin->state |
				WINDOWSIMULATING, VINTEGER);
			(void)setval((INTBIG)schemwin, VWINDOWPART, "charhandler",
				(INTBIG)charhandlerschem, VADDRESS);
			endobjectchange((INTBIG)schemwin, VWINDOWPART);
		}
	}
	return(0);
}

/*
 * Routine to stop simulation.
 */
void sim_window_stopsimulation(void)
{
	REGISTER WINDOWPART *schemwin, *wavewin;

	/* disable simulation control in schematics/layout window */
	schemwin = sim_window_findschematics();
	if (schemwin != NOWINDOWPART)
	{
		startobjectchange((INTBIG)schemwin, VWINDOWPART);
		(void)setval((INTBIG)schemwin, VWINDOWPART, "state", schemwin->state &
			~WINDOWSIMULATING, VINTEGER);
		(void)setval((INTBIG)schemwin, VWINDOWPART, "charhandler",
			(INTBIG)DEFAULTCHARHANDLER, VADDRESS);
		endobjectchange((INTBIG)schemwin, VWINDOWPART);
	}

	/* disable waveform window */
	wavewin = sim_window_findwaveform();
	if (wavewin != NOWINDOWPART)
	{
		startobjectchange((INTBIG)wavewin, VWINDOWPART);
		(void)setval((INTBIG)wavewin, VWINDOWPART, "state", wavewin->state &
			~WINDOWSIMULATING, VINTEGER);
		endobjectchange((INTBIG)wavewin, VWINDOWPART);
	}
}

/*
 * routine that returns:
 *  0                                    if there is no active simulation
 *  SIMWINDOWWAVEFORM                    if simulating and have a waveform window
 *  SIMWINDOWSCHEMATIC                   if simulating and have a schematics window
 *  SIMWINDOWWAVEFORM|SIMWINDOWSCHEMATIC if simulating and have both windows
 * If there is active simulation, the facet being simulated is returned in "np"
 */
INTBIG sim_window_isactive(NODEPROTO **np)
{
	REGISTER INTBIG activity;
	REGISTER WINDOWPART *wavewin, *schemwin;

	*np = NONODEPROTO;
	activity = 0;
	schemwin = sim_window_findschematics();
	if (schemwin != NOWINDOWPART)
	{
		activity |= SIMWINDOWSCHEMATIC;
		*np = schemwin->curnodeproto;
	}
	wavewin = sim_window_findwaveform();
	if (wavewin != NOWINDOWPART && (wavewin->state&WINDOWSIMULATING) != 0)
	{
		activity |= SIMWINDOWWAVEFORM;
		*np = wavewin->curnodeproto;
	}
	return(activity);
}

/*
 * this procedure redraws everything, including the waveforms.
 */
void sim_window_redraw(void)
{
	REGISTER WINDOWPART *wavewin;

	/* get the waveform window */
	wavewin = sim_window_findwaveform();
	if (wavewin == NOWINDOWPART) return;
	sim_window_redisphandler(wavewin);
}

/*
 * routine that changes the number of trace lines in the simulation window to "count"
 */
void sim_window_setlines(INTBIG count)
{
	if (count <= 0) count = 1;
	sim_window_lines = count;
}

/*
 * routine that returns the number of signals in the simulation window
 */
INTBIG sim_window_getlines(void)
{
	return(sim_window_lines);
}

/*
 * routine that changes the number of visible trace lines in the simulation window to "count"
 */
void sim_window_setvislines(INTBIG count)
{
	if (count <= 0) count = 1;
	sim_window_vislines = count;
}

/*
 * routine that returns the number of visible signals in the simulation window
 */
INTBIG sim_window_getvislines(void)
{
	return(sim_window_vislines);
}

/*
 * routine that changes the top visible trace line in the simulation window to "count"
 */
void sim_window_settopline(INTBIG top)
{
	if (top < 0) top = 0;
	sim_window_topline = top;
}

/*
 * routine that returns the top visible signal in the simulation window.
 */
INTBIG sim_window_gettopline(void)
{
	return(sim_window_topline);
}

/********************************* CONTROL OF TRACES *********************************/

/*
 * routine to create a new trace to be displayed in window location
 * "position".  Its name is set to "name", and "userdata" is
 * associated with it.  A pointer to the trace is returned (zero on error).
 */
INTBIG sim_window_newtrace(INTBIG position, char *name, INTBIG userdata)
{
	REGISTER DISCHANNEL *tr, *otr, *lasttr;
	REGISTER INTBIG len, olen, i;

	/* create a new trace */
	tr = sim_window_allocdischannel();
	if (tr == NODISCHANNEL) return(0);
	tr->nodeptr = userdata;

	/* set name and convert if it has array index form */
	(void)allocstring(&tr->name, name, sim_aid->cluster);
	(void)allocstring(&tr->origname, name, sim_aid->cluster);
	len = strlen(tr->name) - 1;
	if (tr->name[len] == '_')
	{
		olen = len;
		while (len > 0 && isdigit(tr->name[len-1])) len--;
		if (len != olen && tr->name[len-1] == '_')
		{
			/* found the form "name_INDEX_" */
			tr->name[len-1] = '[';
			tr->name[strlen(tr->name)-1] = ']';
		}
	}

	/* put in active list */
	lasttr = NODISCHANNEL;
	for(otr = sim_window_plot; otr != NODISCHANNEL; otr = otr->nextdischannel)
		lasttr = otr;
	sim_window_addtrace(tr, lasttr);

	if (position >= 0) tr->position = position; else
	{
		for(i=0, otr = sim_window_plot; otr != NODISCHANNEL; otr = otr->nextdischannel)
			if ((otr->flags&TRACENOTDRAWN) == 0) otr->position = i++;
	}

	/* save the new configuration */
	sim_window_savesignalorder();

	return((INTBIG)tr);
}

/*
 * Routine to open/close a bus.  Returns nonzero on error.
 */
INTSML sim_window_buscommand(void)
{
	REGISTER DISCHANNEL *tr;
	REGISTER INTBIG i, count;
	REGISTER INTSML err;
	INTBIG *tracelist;

	if (sim_window_highlightedtracecount == 0) return(1);
	if (sim_window_highlightedtracecount == 1)
	{
		tr = sim_window_highlightedtraces[0];
		switch (tr->flags&TRACETYPE)
		{
			case TRACEISDIGITAL:
			case TRACEISANALOG:
				ttyputerr(_("Must select multiple signals to pack into a bus"));
				break;
			case TRACEISBUS:
				ttyputmsg(_("Double-click the bus name to open it"));
				break;
			case TRACEISOPENBUS:
				ttyputmsg(_("Double-click the bus name to close it"));
				break;
		}
		return(1);
	}

	/* copy highlighted traces to an array and remove highlighting */
	count = sim_window_highlightedtracecount;
	tracelist = (INTBIG *)emalloc(count * SIZEOFINTBIG, sim_aid->cluster);
	if (tracelist == 0) return(1);
	for(i=0; i<count; i++)
		tracelist[i] = (INTBIG)sim_window_highlightedtraces[i];
	sim_window_cleartracehighlight();

	/* make the bus */
	err = sim_window_makebus(count, tracelist);
	efree((char *)tracelist);
	if (err != 0) return(1);

	/* redisplay */
	sim_window_redraw();

	/* save the new configuration */
	sim_window_savesignalorder();
	return(0);
}

INTSML sim_window_makebus(INTBIG count, INTBIG *traces)
{
	REGISTER DISCHANNEL *tr, *lasttr, *newtr, **tracelist;
	REGISTER INTBIG i, flags;

	/* check for errors */
	tracelist = (DISCHANNEL **)traces;
	for(i=0; i<count; i++)
	{
		flags = tracelist[i]->flags & TRACETYPE;
		if (flags == TRACEISBUS || flags == TRACEISOPENBUS)
		{
			ttyputerr(_("Cannot place a bus into another bus"));
			return(1);
		}
		if (tracelist[i]->buschannel != NODISCHANNEL)
		{
			ttyputerr(_("Some of these signals are already in a bus"));
			return(1);
		}
	}
	if (count > 32)
	{
		ttyputerr(_("Can only handle busses that are 32 bits wide"));
		return(1);
	}

	/* find place to insert bus */
	lasttr = NODISCHANNEL;
	for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
	{
		for(i=0; i<count; i++)
			if (tracelist[i] == tr) break;
		if (i < count) break;
		lasttr = tr;
	}

	/* create a new trace that combines all of these */
	newtr = sim_window_allocdischannel();
	if (newtr == NODISCHANNEL) return(1);
	(void)allocstring(&newtr->name, sim_window_namesignals(count, tracelist),
		sim_aid->cluster);
	(void)allocstring(&newtr->origname, newtr->name, sim_aid->cluster);
	newtr->nodeptr = 0;
	newtr->flags = TRACEISBUS;

	/* insert it after "lasttr" */
	sim_window_addtrace(newtr, lasttr);

	/* flag the signals as part of this bus */
	for(i=0; i<count; i++)
	{
		tr = tracelist[i];
		tr->buschannel = newtr;
		tr->flags |= TRACENOTDRAWN;
	}

	/* renumber the lines in the display */
	for(i=0, tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
		if ((tr->flags&TRACENOTDRAWN) == 0) tr->position = i++;
	return(0);
}

char *sim_window_namesignals(INTBIG count, DISCHANNEL **list)
{
	REGISTER INTBIG i, j, rootlen;
	REGISTER DISCHANNEL *tr, *ptr;

	/* look for common root */
	rootlen = 0;
	for(i=1; i<count; i++)
	{
		ptr = list[i-1];
		tr = list[i];
		for(j=0; ; j++)
		{
			if (ptr->name[j] == 0 || tr->name[0] == 0) break;
			if (ptr->name[j] != tr->name[j]) break;
		}
		if (i == 1 || j < rootlen) rootlen = j;
	}

	/* if there is a common root, make sure it is proper */
	if (rootlen > 0)
	{
		tr = list[0];
		if (tr->name[rootlen-1] != '[')
		{
			for(i=0; i<count; i++)
			{
				tr = list[i];
				if (isdigit(tr->name[rootlen]) == 0) break;
			}
			if (i < count) rootlen = 0;
		}
	}

	/* if there is a common root, use it */
	if (rootlen > 0)
	{
		(void)initinfstr();
		tr = list[0];
		for(j=0; j<rootlen; j++) (void)addtoinfstr(tr->name[j]);

		for(i=0; i<count; i++)
		{
			tr = list[i];
			if (i != 0) (void)addtoinfstr(',');
			for(j=rootlen; tr->name[j] != ']' && tr->name[j] != 0; j++)
				(void)addtoinfstr(tr->name[j]);
		}
		if (tr->name[rootlen-1] == '[') (void)addtoinfstr(']');
		return(returninfstr());
	}

	/* no common root: concatenate all signal names */
	(void)initinfstr();
	for(i=0; i<count; i++)
	{
		if (i != 0) (void)addtoinfstr(',');
		tr = list[i];
		(void)addstringtoinfstr(tr->name);
	}
	return(returninfstr());
}

void sim_window_savesignalorder(void)
{
	REGISTER DISCHANNEL *tr, *otr;
	NODEPROTO *np;
	char **strings;
	INTBIG count, i, j, k;

	if (sim_window_isactive(&np) == 0) return;
	for(count = 0, tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
		if ((tr->flags&TRACENOTDRAWN) == 0 && tr->buschannel == NODISCHANNEL) count++;
	if (count == 0)
	{
		if (getvalkey((INTBIG)np, VNODEPROTO, VSTRING|VISARRAY, sim_window_signal_order) != NOVARIABLE)
			delvalkey((INTBIG)np, VNODEPROTO, sim_window_signal_order);
	} else
	{
		strings = (char **)emalloc(count * (sizeof (char *)), sim_aid->cluster);
		for(i=0, j=count, tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel, i++)
		{
			if ((tr->flags&TRACENOTDRAWN) != 0 || tr->buschannel != NODISCHANNEL) continue;
			if ((tr->flags&TRACETYPE) == TRACEISBUS || (tr->flags&TRACETYPE) == TRACEISOPENBUS)
			{
				/* construct name for this bus */
				(void)initinfstr();
				k = 0;
				for(otr = sim_window_plot; otr != NODISCHANNEL; otr = otr->nextdischannel)
				{
					if (otr->buschannel != tr) continue;
					if (k != 0) (void)addtoinfstr('\t');
					(void)addstringtoinfstr(otr->origname);
					k++;
				}
				allocstring(&strings[--j], returninfstr(), sim_aid->cluster);
			} else
			{
				allocstring(&strings[--j], tr->origname, sim_aid->cluster);
			}
		}
		(void)setvalkey((INTBIG)np, VNODEPROTO, sim_window_signal_order, (INTBIG)strings,
			VSTRING|VISARRAY|(count<<VLENGTHSH));
		for(i=0; i<count; i++) efree((char *)strings[i]);
		efree((char *)strings);
	}
}

/*
 * routine to load trace "tri" with "count" digital events.  The events occur
 * at time "time" and have state "state".  "State" is encoded with a
 * level in the upper 8 bits (LOGIC_LOW, LOGIC_HIGH, or LOGIC_X) and a
 * strength in the lower 8 bits (OFF_STRENGTH, NODE_STRENGTH, GATE_STRENGTH,
 * or VDD_STRENGTH).
 */
void sim_window_loaddigtrace(INTBIG tri, INTBIG count, float *time, INTSML *state)
{
	REGISTER INTBIG wanted, i;
	REGISTER DISCHANNEL *tr;

	/* ensure space in the arrays */
	tr = (DISCHANNEL *)tri;
	if (tr == NODISCHANNEL || tr == 0) return;
	if (count > tr->timelimit)
	{
		if (tr->timelimit > 0)
		{
			efree((char *)tr->timearray);
			tr->timelimit = 0;
		}
		wanted = count + 10;
		tr->timearray = (float *)emalloc(wanted * (sizeof (float)), sim_aid->cluster);
		if (tr->timearray == 0) return;
		tr->timelimit = wanted;
	}
	if (count > tr->statelimit)
	{
		if (tr->statelimit > 0)
		{
			efree((char *)tr->statearray);
			tr->statelimit = 0;
		}
		wanted = count + 10;
		tr->statearray = (INTBIG *)emalloc(wanted * SIZEOFINTBIG, sim_aid->cluster);
		if (tr->statearray == 0) return;
		tr->statelimit = wanted;
	}

	/* load the data */
	for(i=0; i<count; i++)
	{
		tr->timearray[i] = time[i];
		tr->statearray[i] = state[i];
	}
	tr->flags = (tr->flags & ~TRACETYPE) | TRACEISDIGITAL;
	tr->numsteps = count;
}

/*
 * routine to load trace "tri" with "count" analog events.  The events occur
 * at time "time" and have value "value".
 */
void sim_window_loadanatrace(INTBIG tri, INTBIG count, float *time, float *value)
{
	REGISTER INTBIG wanted, i;
	REGISTER DISCHANNEL *tr;

	/* ensure space in the arrays */
	tr = (DISCHANNEL *)tri;
	if (tr == NODISCHANNEL || tr == 0) return;
	if (count > tr->timelimit)
	{
		if (tr->timelimit > 0)
		{
			efree((char *)tr->timearray);
			tr->timelimit = 0;
		}
		wanted = count + 10;
		tr->timearray = (float *)emalloc(wanted * (sizeof (float)), sim_aid->cluster);
		if (tr->timearray == 0) return;
		tr->timelimit = wanted;
	}
	if (count > tr->valuelimit)
	{
		if (tr->valuelimit > 0)
		{
			efree((char *)tr->valuearray);
			tr->valuelimit = 0;
		}
		wanted = count + 10;
		tr->valuearray = (float *)emalloc(wanted * (sizeof (float)), sim_aid->cluster);
		if (tr->valuearray == 0) return;
		tr->valuelimit = wanted;
	}

	/* load the data */
	for(i=0; i<count; i++)
	{
		tr->timearray[i] = time[i];
		tr->valuearray[i] = value[i];
	}
	tr->flags = (tr->flags & ~TRACETYPE) | TRACEISANALOG;
	tr->color = RED;
	tr->numsteps = count;
}

/*
 * Sets the analog range to "low"-to-"high".  This is the displayed vertical range.
 */
void sim_window_setanarange(float low, float high)
{
	sim_window_anadislow = low;
	sim_window_anadishigh = high;
	sim_window_anadisrange = high - low;
}

/*
 * Gets the analog range into "low"-to-"high".  This is the displayed vertical range.
 */
void sim_window_getanarange(float *low, float *high)
{
	*low = sim_window_anadislow;
	*high = sim_window_anadishigh;
}

/*
 * Sets the analog extent to "low"-to-"high".  This is the absolute extent of the data.
 */
void sim_window_setanaextents(float low, float high)
{
	sim_window_analow = low;
	sim_window_anahigh = high;
	sim_window_anarange = high - low;
}

/*
 * Gets the analog extent in "low"-to-"high".  This is the absolute extent of the data.
 */
void sim_window_getanaextents(float *low, float *high)
{
	*low = sim_window_analow;
	*high = sim_window_anahigh;
}

void sim_window_settraceline(INTBIG tri, INTBIG line)
{
	REGISTER DISCHANNEL *tr;

	tr = (DISCHANNEL *)tri;
	if (tr == NODISCHANNEL || tr == 0) return;
	tr->position = line;
}

/*
 * routine to delete the trace "tri"
 */
void sim_window_killtrace(INTBIG tri)
{
	REGISTER DISCHANNEL *tr, *str, *ltr, *otr;
	REGISTER INTBIG i;

	tr = (DISCHANNEL *)tri;
	if (tr == NODISCHANNEL || tr == 0) return;
	ltr = NODISCHANNEL;
	for(str = sim_window_plot; str != NODISCHANNEL; str = str->nextdischannel)
	{
		if (str == tr) break;
		ltr = str;
	}
	if (ltr == NODISCHANNEL) sim_window_plot = tr->nextdischannel; else
		ltr->nextdischannel = tr->nextdischannel;
	tr->nextdischannel = sim_window_plotfree;
	sim_window_plotfree = tr;

	/* if a bus was deleted, make it's individual signals be visible and independent */
	if ((tr->flags&TRACETYPE) == TRACEISBUS || (tr->flags&TRACETYPE) == TRACEISOPENBUS)
	{
		for(otr = sim_window_plot; otr != NODISCHANNEL; otr = otr->nextdischannel)
		{
			if (otr->buschannel != tr) continue;
			otr->buschannel = NODISCHANNEL;
			otr->flags &= ~TRACENOTDRAWN;
		}
	}
	efree(tr->name);
	efree(tr->origname);

	/* renumber the lines in the display */
	for(i=0, tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
		if ((tr->flags&TRACENOTDRAWN) == 0) tr->position = i++;

	/* save the new configuration */
	sim_window_savesignalorder();
}

/*
 * routine to delete all traces
 */
void sim_window_killalltraces(void)
{
	REGISTER DISCHANNEL *tr, *nexttr;

	/* return all plot channels to the free list */
	for(tr = sim_window_plot; tr != NODISCHANNEL; tr = nexttr)
	{
		nexttr = tr->nextdischannel;
		tr->nextdischannel = sim_window_plotfree;
		sim_window_plotfree = tr;
		efree(tr->name);
		efree(tr->origname);
	}
	sim_window_plot = NODISCHANNEL;
	sim_window_highlightedtracecount = 0;

	/* save the new configuration */
	sim_window_savesignalorder();
}

/********************************* TRACE INFORMATION *********************************/

INTBIG *sim_window_getbustraces(INTBIG tri)
{
	REGISTER DISCHANNEL *tr, *otr;
	REGISTER INTBIG count, newtotal, i;
	REGISTER DISCHANNEL **newbustraces;
	static INTBIG nullbustraces[1];

	tr = (DISCHANNEL *)tri;
	count = 0;
	for(otr = sim_window_plot; otr != NODISCHANNEL; otr = otr->nextdischannel)
	{
		if (otr->buschannel != tr) continue;
		if (count+1 >= sim_window_bustracetotal)
		{
			newtotal = sim_window_bustracetotal*2;
			if (newtotal == 0) newtotal = 5;
			if (count+1 > newtotal) newtotal = count+1;
			newbustraces = (DISCHANNEL **)emalloc(newtotal * (sizeof (DISCHANNEL *)),
				sim_aid->cluster);
			if (newbustraces == 0) return(0);
			for(i=0; i<count; i++)
				newbustraces[i] = sim_window_bustraces[i];
			if (sim_window_bustracetotal > 0) efree((char *)sim_window_bustraces);
			sim_window_bustraces = newbustraces;
			sim_window_bustracetotal = newtotal;
		}
		sim_window_bustraces[count++] = otr;
		sim_window_bustraces[count] = 0;
	}
	if (count == 0)
	{
		nullbustraces[0] = 0;
		return(nullbustraces);
	}
	return((INTBIG *)sim_window_bustraces);
}

/*
 * routine to return the line number associated with trace "tri"
 */
INTBIG sim_window_gettraceline(INTBIG tri)
{
	REGISTER DISCHANNEL *tr;

	tr = (DISCHANNEL *)tri;
	if (tr == NODISCHANNEL || tr == 0) return(0);
	return(tr->position);
}

/*
 * routine to return the name of trace "tri".
 */
char *sim_window_gettracename(INTBIG tri)
{
	REGISTER DISCHANNEL *tr;

	tr = (DISCHANNEL *)tri;
	if (tr == NODISCHANNEL || tr == 0) return("");
	return(tr->origname);
}

/*
 * routine to return the "user" data associated with trace "tri"
 */
INTBIG sim_window_gettracedata(INTBIG tri)
{
	REGISTER DISCHANNEL *tr;

	tr = (DISCHANNEL *)tri;
	if (tr == NODISCHANNEL || tr == 0) return(0);
	return(tr->nodeptr);
}

float sim_window_getanatracevalue(INTBIG tri, float time)
{
	REGISTER DISCHANNEL *tr;
	REGISTER INTBIG j;
	REGISTER float lasttime, lastvalue, thistime, thisvalue, v;

	tr = (DISCHANNEL *)tri;
	if (tr == NODISCHANNEL || tr == 0) return(0.0);

	/* must be an analog trace */
	if ((tr->flags&TRACETYPE) != TRACEISANALOG) return(0.0);

	for(j=0; j<tr->numsteps; j++)
	{
		thistime = tr->timearray[j];
		thisvalue = tr->valuearray[j] * sim_window_anarange + sim_window_analow;
		if (thistime == time) return(thisvalue);
		if (time > thistime)
		{
			lasttime = thistime;
			lastvalue = thisvalue;
			continue;
		}
		if (j == 0) return(thisvalue);
		v = lastvalue + (time-lasttime) * (thisvalue-lastvalue) / (thistime-lasttime);
		return(v);
	}

	return(lastvalue);
}

INTBIG sim_window_findtrace(char *name)
{
	REGISTER DISCHANNEL *tr;
	REGISTER char *tempname;

	for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
	{
		if ((tr->flags&TRACENOTDRAWN) != 0) continue;
		if (namesame(tr->name, name) == 0) return((INTBIG)tr);
	}

	/* keywords get the letters "NV" added to the end */
	(void)initinfstr();
	(void)addstringtoinfstr(name);
	(void)addstringtoinfstr("NV");
	tempname = returninfstr();
	for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
	{
		if ((tr->flags&TRACENOTDRAWN) != 0) continue;
		if (namesame(tr->name, tempname) == 0) return((INTBIG)tr);
	}

	/* handle "v(" and "l(" prefixes of HSPICE (see also "dbtext.c:getcomplexnetwork()") */
	for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
	{
		if ((tr->flags&TRACENOTDRAWN) != 0) continue;
		if ((tr->name[0] == 'l' || tr->name[0] == 'v') && tr->name[1] == '(')
		{
			if (namesame(&tr->name[2], name) == 0) return((INTBIG)tr);
		}
	}

	/* name not found in simulator */
	return(0);
}

void sim_window_inittraceloop(void)
{
	sim_window_loop = sim_window_plot;
}

void sim_window_inittraceloop2(void)
{
	sim_window_loop2 = sim_window_plot;
}

INTBIG sim_window_nexttraceloop(void)
{
	REGISTER DISCHANNEL *tr;

	tr = sim_window_loop;
	if (tr == NODISCHANNEL) return(0);
	sim_window_loop = tr->nextdischannel;
	return((INTBIG)tr);
}

INTBIG sim_window_nexttraceloop2(void)
{
	REGISTER DISCHANNEL *tr;

	tr = sim_window_loop2;
	if (tr == NODISCHANNEL) return(0);
	sim_window_loop2 = tr->nextdischannel;
	return((INTBIG)tr);
}

/********************************* TRACE HIGHLIGHTING *********************************/

/*
 * Routine to clear all highlighting in the waveform window.
 */
void sim_window_cleartracehighlight(void)
{
	REGISTER WINDOWPART *wavewin;
	REGISTER INTBIG i;

	/* get the waveform window */
	wavewin = sim_window_findwaveform();
	if (wavewin == NOWINDOWPART) return;
	sim_window_setviewport(wavewin);

	/* erase highlighting */
	for(i=0; i<sim_window_highlightedtracecount; i++)
		sim_window_plottrace(sim_window_highlightedtraces[i], wavewin, 2);
	sim_window_highlightedtracecount = 0;
}

/*
 * routine that adds trace "tri" to the highlighted traces in the waveform window
 */
void sim_window_addhighlighttrace(INTBIG tri)
{
	REGISTER DISCHANNEL *tr, **newtraces;
	REGISTER WINDOWPART *wavewin;
	REGISTER INTBIG newtotal, i;

	tr = (DISCHANNEL *)tri;

	/* get the waveform window */
	wavewin = sim_window_findwaveform();
	if (wavewin == NOWINDOWPART) return;
	sim_window_setviewport(wavewin);

	/* erase highlighting */
	for(i=0; i<sim_window_highlightedtracecount; i++)
		sim_window_plottrace(sim_window_highlightedtraces[i], wavewin, 2);

	/* add this signal to highlighting */
	if (sim_window_highlightedtracecount+1 >= sim_window_highlightedtracetotal)
	{
		newtotal = sim_window_highlightedtracetotal * 2;
		if (newtotal == 0) newtotal = 2;
		if (sim_window_highlightedtracecount+1 > newtotal)
			newtotal = sim_window_highlightedtracecount+1;
		newtraces = (DISCHANNEL **)emalloc(newtotal * (sizeof (DISCHANNEL *)), sim_aid->cluster);
		if (newtraces == 0) return;
		for(i=0; i<sim_window_highlightedtracecount; i++)
			newtraces[i] = sim_window_highlightedtraces[i];
		if (sim_window_highlightedtracetotal > 0) efree((char *)sim_window_highlightedtraces);
		sim_window_highlightedtraces = newtraces;
		sim_window_highlightedtracetotal = newtotal;
	}
	sim_window_highlightedtraces[sim_window_highlightedtracecount++] = tr;
	sim_window_highlightedtraces[sim_window_highlightedtracecount] = 0;

	/* show highlighting */
	for(i=0; i<sim_window_highlightedtracecount; i++)
		sim_window_plottrace(sim_window_highlightedtraces[i], wavewin, 1);
}

/*
 * routine that removes trace "tri" from the highlighted traces in the waveform window
 */
void sim_window_deletehighlighttrace(INTBIG tri)
{
	REGISTER DISCHANNEL *tr;
	REGISTER WINDOWPART *wavewin;
	REGISTER INTBIG k, j, i;

	tr = (DISCHANNEL *)tri;

	/* get the waveform window */
	wavewin = sim_window_findwaveform();
	if (wavewin == NOWINDOWPART) return;
	sim_window_setviewport(wavewin);

	/* find this signal in the highlighting */
	for(j=0; j<sim_window_highlightedtracecount; j++)
		if (tr == sim_window_highlightedtraces[j]) break;
	if (j < sim_window_highlightedtracecount)
	{
		/* erase highlighting */
		for(i=0; i<sim_window_highlightedtracecount; i++)
			sim_window_plottrace(sim_window_highlightedtraces[i], wavewin, 2);

		/* remove the signal from the list */
		k = 0;
		for(i=0; i<sim_window_highlightedtracecount; i++)
		{
			if (i == j) continue;
			sim_window_highlightedtraces[k++] = sim_window_highlightedtraces[i];
		}
		sim_window_highlightedtracecount = k;

		/* show highlighting */
		for(i=0; i<sim_window_highlightedtracecount; i++)
			sim_window_plottrace(sim_window_highlightedtraces[i], wavewin, 1);
	}
}

/*
 * routine that returns the highlighted trace in the waveform window
 */
INTBIG sim_window_gethighlighttrace(void)
{
	if (sim_window_highlightedtracecount == 0) return(0);
	return((INTBIG)sim_window_highlightedtraces[0]);
}

/*
 * routine that returns a 0-terminated list of highlighted traces in the waveform window
 */
INTBIG *sim_window_gethighlighttraces(void)
{
	static INTBIG nothing[1];

	nothing[0] = 0;
	if (sim_window_highlightedtracecount == 0) return(nothing);
	return((INTBIG *)sim_window_highlightedtraces);
}

/*
 * Routine to update associated layout windows when the main cursor changes
 */
void sim_window_updatelayoutwindow(void)
{
	REGISTER DISCHANNEL *tr;
	REGISTER NETWORK *net;
	REGISTER ARCINST *ai;
	REGISTER INTBIG j, fx, fy, tx, ty, windowstate, texture, strength;
	REGISTER VARIABLE *var;
	REGISTER NODEPROTO *simnt;
	REGISTER WINDOWPART *schemwin;
	static POLYGON *poly = NOPOLYGON;
	void us_wanttodraw(INTBIG, INTBIG, INTBIG, INTBIG, WINDOWPART*, GRAPHICS*, INTSML);

	/* make sure there is a layout/schematic window being simulated */
	schemwin = sim_window_findschematics();
	if (schemwin == NOWINDOWPART) return;
	simnt = schemwin->curnodeproto;

	/* get window state information */
	var = getvalkey((INTBIG)sim_aid, VAID, VINTEGER, sim_window_state);
	if (var == NOVARIABLE) windowstate = 0; else windowstate = var->addr;

	/* reset all values on networks */
	for(net = simnt->firstnetwork; net != NONETWORK; net = net->nextnetwork)
		net->temp1 = -1;

	/* assign values from simulation window traces to networks */
	for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
	{
		if ((tr->flags&TRACENOTDRAWN) != 0) continue;

		/* can only handle digital waveforms right now */
		if ((tr->flags&TRACETYPE) != TRACEISDIGITAL) continue;

		/* find the proper data for the time of the main cursor */
		for(j=1; j<tr->numsteps; j++)
			if (sim_window_maincursor < tr->timearray[j]) break;
		j--;

		/* set simulation value on the network in the associated layout/schematic window */
		net = getcomplexnetwork(tr->name, simnt);
		if (net != NONETWORK && j < tr->numsteps) net->temp1 = tr->statearray[j];
	}

	/* redraw all arcs in the layout/schematic window */
	sim_window_desc.bits = LAYERA;
	for(ai = simnt->firstarcinst; ai != NOARCINST; ai = ai->nextarcinst)
	{
		net = ai->network;
		if (net == NONETWORK) continue;
		if (net->temp1 == -1) continue;

		/* get extent of arc */
		fx = ai->end[0].xpos;   fy = ai->end[0].ypos;
		tx = ai->end[1].xpos;   ty = ai->end[1].ypos;
		if (ai->proto == sch_wirearc)
		{
			/* for schematic wires, ask technology about drawing so negating bubbles work */
			if (poly == NOPOLYGON) poly = allocstaticpolygon(4, sim_aid->cluster);
			(void)arcpolys(ai, NOWINDOWPART);
			shapearcpoly(ai, 0, poly);
			fx = poly->xv[0];   fy = poly->yv[0];
			tx = poly->xv[1];   ty = poly->yv[1];
		}

		if ((windowstate&FULLSTATE) != 0)
		{
			/* 12-state display: determine trace texture */
			/* determine trace texture */
			strength = net->temp1 & 0377;
			if (strength == 0) texture = -1; else
				if (strength <= NODE_STRENGTH) texture = 1; else
					if (strength <= GATE_STRENGTH) texture = 0; else
						texture = 2;

			/* reset background arc if trace is textured */
			if (texture != 0)
			{
				sim_window_desc.col = ALLOFF;
				us_wanttodraw(fx, fy, tx, ty, schemwin, &sim_window_desc, 0);
				if (texture < 0) texture = 0;
			}

			/* determine trace color */
			switch (net->temp1 >> 8)
			{
				case LOGIC_LOW:  sim_window_desc.col = (INTSML)sim_colorlevellow;     break;
				case LOGIC_X:    sim_window_desc.col = (INTSML)sim_colorlevelundef;   break;
				case LOGIC_HIGH: sim_window_desc.col = (INTSML)sim_colorlevelhigh;    break;
			}
		} else
		{
			/* 2-state display */
			texture = 0;
			if ((net->temp1 >> 8) == LOGIC_HIGH) sim_window_desc.col = (INTSML)sim_colorstrengthgate; else
				sim_window_desc.col = (INTSML)sim_colorstrengthoff;
		}

		/* draw the trace on the arc */
		us_wanttodraw(fx, fy, tx, ty, schemwin, &sim_window_desc, (INTSML)texture);
	}
}

/********************************* TIME AND CURSOR CONTROL *********************************/

/*
 * routine to set the time for the "main" cursor in the simulation window
 */
void sim_window_setmaincursor(float time)
{
	REGISTER WINDOWPART *wavewin;

	/* get the waveform window */
	wavewin = sim_window_findwaveform();
	if (wavewin == NOWINDOWPART) return;

	sim_window_maincursor = time;
	sim_window_updatelayoutwindow();
	sim_window_drawcursors(wavewin);
}

/*
 * routine to return the time of the "main" cursor in the simulation window
 */
float sim_window_getmaincursor(void)
{
	return(sim_window_maincursor);
}

/*
 * routine to set the time for the "extension" cursor in the simulation window
 */
void sim_window_setextensioncursor(float time)
{
	REGISTER WINDOWPART *wavewin;

	/* get the waveform window */
	wavewin = sim_window_findwaveform();
	if (wavewin == NOWINDOWPART) return;

	sim_window_extensioncursor = time;
	sim_window_drawcursors(wavewin);
}

/*
 * routine to return the time of the "extension" cursor in the simulation window
 */
float sim_window_getextensioncursor(void)
{
	return(sim_window_extensioncursor);
}

/*
 * routine to set the simulation window to run from time "min" to time "max"
 */
void sim_window_settimerange(float min, float max)
{
	sim_window_mintime = min;
	sim_window_maxtime = max;
}

/*
 * routine to return the range of time displayed in the simulation window
 */
void sim_window_gettimerange(float *min, float *max)
{
	*min = sim_window_mintime;
	*max = sim_window_maxtime;
}

/*
 * routine to return the range of time that is in the data
 */
void sim_window_gettimeextents(float *min, float *max)
{
	REGISTER DISCHANNEL *tr;
	float time;
	REGISTER INTBIG j;

	/* initially enclose the cursors */
	*min = sim_window_maincursor;
	*max = sim_window_extensioncursor;
	if (*max < *min)
	{
		*min = sim_window_extensioncursor;
		*max = sim_window_maincursor;
	}

	/* also enclose all of the data */
	for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
	{
		if ((tr->flags&TRACENOTDRAWN) != 0) continue;
		for(j=0; j<tr->numsteps; j++)
		{
			time = tr->timearray[j];
			if (time < *min) *min = time;
			if (time > *max) *max = time;
		}
	}
}

/********************************* WINDOW METHODS *********************************/

INTSML sim_window_eachwdown(INTBIG x, INTBIG y)
{
	REGISTER WINDOWPART *wavewin;

	/* get the waveform window */
	wavewin = sim_window_findwaveform();
	if (wavewin == NOWINDOWPART) return(0);

	sim_window_scaletowindow(wavewin, &x, &y);
	sim_window_maincursor = sim_window_xpostotime(x);
	if (sim_window_maincursor < sim_window_mintime)
		sim_window_maincursor = sim_window_mintime;
	if (sim_window_maincursor > sim_window_maxtime)
		sim_window_maincursor = sim_window_maxtime;
	sim_window_drawcursors(wavewin);
	sim_window_updatelayoutwindow();
	return(0);
}

INTSML sim_window_eachbdown(INTBIG x, INTBIG y)
{
	REGISTER WINDOWPART *wavewin;

	/* get the waveform window */
	wavewin = sim_window_findwaveform();
	if (wavewin == NOWINDOWPART) return(0);

	sim_window_scaletowindow(wavewin, &x, &y);
	sim_window_extensioncursor = sim_window_xpostotime(x);
	if (sim_window_extensioncursor < sim_window_mintime)
		sim_window_extensioncursor = sim_window_mintime;
	if (sim_window_extensioncursor > sim_window_maxtime)
		sim_window_extensioncursor = sim_window_maxtime;
	sim_window_drawcursors(wavewin);
	return(0);
}

/*
 * Routine for drawing highlight around dragged trace name.
 */
void sim_window_drawtracedrag(WINDOWPART *wavewin, INTSML on)
{
	INTBIG y_pos, temp_trace_spacing, line, lx, hx, ly, hy;

	/* compute positions */
	temp_trace_spacing = 540 / sim_window_vislines;
	line = sim_window_tracedragtr->position - sim_window_topline;
	if (line < 0 || line >= sim_window_vislines) return;
	y_pos = (sim_window_vislines - 1 - line) * temp_trace_spacing +
		31 + sim_window_tracedragyoff;
	ly = y_pos + sim_window_tracedragtr->nameoff + 1;
	hy = ly + sim_window_txthei;
	lx = 0;
	hx = 112;

	sim_window_setmask(LAYERH);
	if (on == 0) sim_window_setcolor(ALLOFF); else
		sim_window_setcolor(HIGHLIT);
	sim_window_moveto(lx, ly);
	sim_window_drawto(wavewin, hx, ly);
	sim_window_drawto(wavewin, hx, hy);
	sim_window_drawto(wavewin, lx, hy);
	sim_window_drawto(wavewin, lx, ly);
}

/*
 * Routine for tracking the drag of signal names in the waveform window.
 */
INTSML sim_window_eachdown(INTBIG x, INTBIG y)
{
	REGISTER WINDOWPART *wavewin;

	/* get the waveform window */
	wavewin = sim_window_findwaveform();
	if (wavewin == NOWINDOWPART) return(0);

	sim_window_scaletowindow(wavewin, &x, &y);
	sim_window_drawtracedrag(wavewin, 0);
	sim_window_tracedragyoff = y - sim_window_initialy;
	sim_window_drawtracedrag(wavewin, 1);
	return(0);
}

/*
 * Routine for drawing highlight around selected trace names.
 */
void sim_window_drawtraceselect(WINDOWPART *wavewin, INTBIG y1, INTBIG y2, INTSML on)
{
	INTBIG x, ly, hy;

	/* compute positions */
	if (y1 == y2) return;
	ly = mini(y1, y2);
	hy = maxi(y1, y2);
	x = sim_window_initialx;

	sim_window_setmask(LAYERH);
	if (on == 0) sim_window_setcolor(ALLOFF); else
		sim_window_setcolor(HIGHLIT);
	sim_window_moveto(x, ly);
	sim_window_drawto(wavewin, x, hy);
}

/*
 * Routine for tracking the selection of signal names in the waveform window.
 */
INTSML sim_window_seleachdown(INTBIG x, INTBIG y)
{
	REGISTER WINDOWPART *wavewin;
	INTBIG lowy, highy, ind, i, redo, ypos, temp_trace_spacing;
	REGISTER DISCHANNEL *tr;

	/* get the waveform window */
	wavewin = sim_window_findwaveform();
	if (wavewin == NOWINDOWPART) return(0);

	sim_window_scaletowindow(wavewin, &x, &y);

	/* undraw former selection line */
	sim_window_drawtraceselect(wavewin, sim_window_currentdragy, sim_window_initialy, 0);
	if (y < sim_window_initialy)
	{
		lowy = y;    highy = sim_window_initialy;
	} else
	{
		lowy = sim_window_initialy;    highy = y;
	}

	for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
		tr->flags &= ~TRACEWANTSELECT;
	temp_trace_spacing = 540 / sim_window_vislines;
	for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
	{
		if ((tr->flags&TRACENOTDRAWN) != 0) continue;
		ind = tr->position - sim_window_topline;
		ypos = (sim_window_vislines - 1 - ind) * temp_trace_spacing + 31 + tr->nameoff;
		if (ypos > highy || ypos + sim_window_txthei < lowy) continue;
		tr->flags |= TRACEWANTSELECT;
	}

	/* see if what is highlighted has changed */
	redo = 0;
	for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
	{
		if ((tr->flags&TRACENOTDRAWN) != 0) continue;
		for(i=0; i<sim_window_highlightedtracecount; i++)
			if (sim_window_highlightedtraces[i] == tr) break;
		if (i < sim_window_highlightedtracecount)
		{
			/* this signal was highlighted */
			if ((tr->flags&TRACEWANTSELECT) == 0) redo++;
		} else
		{
			/* this signal was not highlighted */
			if ((tr->flags&TRACEWANTSELECT) != 0) redo++;
		}
	}

	/* redo highlighting if it changed */
	if (redo != 0)
	{
		sim_window_cleartracehighlight();
		for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
			if ((tr->flags&TRACEWANTSELECT) != 0) sim_window_addhighlighttrace((INTBIG)tr);
	}

	/* show what is dragged out */
	sim_window_currentdragy = y;
	sim_window_drawtraceselect(wavewin, sim_window_currentdragy, sim_window_initialy, HIGHLIT);
	return(0);
}

/*
 * Routine called when vertical slider parts are clicked.
 */
INTSML sim_window_verticalslider(INTBIG x, INTBIG y)
{
	INTBIG newtopline;

	if (x >= sim_waveformwindow->uselx) return(0);
	switch (sim_waveformsliderpart)
	{
		case 0: /* down arrow clicked */
			if (y > sim_waveformwindow->usely + DISPLAYSLIDERSIZE) return(0);
			if (y < sim_waveformwindow->usely) return(0);
			if (sim_window_topline + sim_window_vislines < sim_window_lines)
			{
				sim_window_topline++;
				sim_window_redisphandler(sim_waveformwindow);
			}
			break;
		case 1: /* below thumb */
			if (y > sim_waveformwindow->thumbly) return(0);
			newtopline = sim_window_topline + sim_window_vislines - 1;
			if (newtopline + sim_window_vislines > sim_window_lines)
				newtopline = sim_window_lines - sim_window_vislines;
			if (sim_window_topline == newtopline) break;
			sim_window_topline = newtopline;
			sim_window_redisphandler(sim_waveformwindow);
			break;
		case 2: /* on thumb (not handled here) */
			break;
		case 3: /* above thumb */
			if (y < sim_waveformwindow->thumbhy) return(0);
			if (sim_window_topline == 0) return(0);
			sim_window_topline -= sim_window_vislines - 1;
			if (sim_window_topline < 0) sim_window_topline = 0;
			sim_window_redisphandler(sim_waveformwindow);
			break;
		case 4: /* up arrow clicked */
			if (y < sim_waveformwindow->usehy - DISPLAYSLIDERSIZE) return(0);
			if (sim_window_topline > 0)
			{
				sim_window_topline--;
				sim_window_redisphandler(sim_waveformwindow);
			}
			break;
	}
	return(0);
}

/*
 * Routine called when horizontal slider parts are clicked.
 */
INTSML sim_window_horizontalslider(INTBIG x, INTBIG y)
{
	float range;

	if (y >= sim_waveformwindow->usely) return(0);
	switch (sim_waveformsliderpart)
	{
		case 0: /* left arrow clicked */
			if (x > sim_waveformwindow->uselx + DISPLAYSLIDERSIZE) return(0);
			if (sim_window_mintime <= 0.0) break;
			range = sim_window_maxtime - sim_window_mintime;
			sim_window_mintime -= range/DISPLAYSMALLSHIFT;
			sim_window_maxtime -= range/DISPLAYSMALLSHIFT;
			if (sim_window_mintime < 0.0)
			{
				sim_window_maxtime -= sim_window_mintime;
				sim_window_mintime = 0.0;
			}
			simals_initialize_simulator(1);
			break;
		case 1: /* left of thumb */
			if (x > sim_waveformwindow->thumblx) return(0);
			if (sim_window_mintime <= 0.0) break;
			range = sim_window_maxtime - sim_window_mintime;
			sim_window_mintime -= range/DISPLAYLARGESHIFT;
			sim_window_maxtime -= range/DISPLAYLARGESHIFT;
			if (sim_window_mintime < 0.0)
			{
				sim_window_maxtime -= sim_window_mintime;
				sim_window_mintime = 0.0;
			}
			simals_initialize_simulator(1);
			break;
		case 2: /* on thumb (not handled here) */
			break;
		case 3: /* right of thumb */
			if (x < sim_waveformwindow->thumbhx) return(0);
			range = sim_window_maxtime - sim_window_mintime;
			sim_window_mintime += range/DISPLAYLARGESHIFT;
			sim_window_maxtime += range/DISPLAYLARGESHIFT;
			simals_initialize_simulator(1);
			break;
		case 4: /* right arrow clicked */
			if (x < sim_waveformwindow->usehx - DISPLAYSLIDERSIZE) return(0);
			range = sim_window_maxtime - sim_window_mintime;
			sim_window_mintime += range/DISPLAYSMALLSHIFT;
			sim_window_maxtime += range/DISPLAYSMALLSHIFT;
			simals_initialize_simulator(1);
			break;
	}
	return(0);
}

void sim_window_buttonhandler(WINDOWPART *w, INTSML button, INTSML inx, INTSML iny)
{
	INTBIG x, y, xpos, ypos, newtop, i, j, ind, temp_trace_spacing, tracelinesmoved;
	float newpos;
	char *par[2];
	REGISTER DISCHANNEL *tr, *otr, *trprev, *otrprev, *nexttr, *addloc,
		*buschain, *endbuschain;

	/* changes to the mouse-wheel are handled by the user interface */
	if (wheelbutton(button) != 0)
	{
		us_buttonhandler(w, button, inx, iny);
		return;
	}
	ttynewcommand();

	/* hit in slider on left */
	if (inx < w->uselx)
	{
		if (iny <= w->usely + DISPLAYSLIDERSIZE)
		{
			/* hit in bottom arrow: go down 1 (may repeat) */
			sim_waveformwindow = w;
			sim_waveformsliderpart = 0;
			trackcursor(0, us_nullup, us_nullvoid, sim_window_verticalslider, us_nullchar,
				us_nullvoid, TRACKNORMAL);
			return;
		}
		if (iny < w->thumbly)
		{
			/* hit below thumb: go down lots (may repeat) */
			sim_waveformwindow = w;
			sim_waveformsliderpart = 1;
			trackcursor(0, us_nullup, us_nullvoid, sim_window_verticalslider, us_nullchar,
				us_nullvoid, TRACKNORMAL);
			return;
		}
		if (iny <= w->thumbhy)
		{
			/* hit on thumb: track thumb motion */
			us_vthumbbegin(iny, w, w->uselx-DISPLAYSLIDERSIZE, w->usely, w->usehy,
				sim_window_lines, 1);
			trackcursor(0, us_nullup, us_nullvoid, us_vthumbdown, us_nullchar,
				us_svthumbdone, TRACKNORMAL); 
			newtop = us_svthumbend();
			if (newtop >= 0)
			{
				sim_window_topline = newtop;
				sim_window_redisphandler(w);
			}
			return;
		}
		if (iny < w->usehy - DISPLAYSLIDERSIZE)
		{
			/* hit above thumb: go up lots (may repeat) */
			sim_waveformwindow = w;
			sim_waveformsliderpart = 3;
			trackcursor(0, us_nullup, us_nullvoid, sim_window_verticalslider, us_nullchar,
				us_nullvoid, TRACKNORMAL);
			return;
		}

		/* hit in up arrow: go up 1 (may repeat) */
		sim_waveformwindow = w;
		sim_waveformsliderpart = 4;
		trackcursor(0, us_nullup, us_nullvoid, sim_window_verticalslider, us_nullchar,
			us_nullvoid, TRACKNORMAL);
		return;
	}

	/* hit in slider on bottom */
	if (iny < w->usely)
	{
		if (inx <= w->uselx + DISPLAYSLIDERSIZE)
		{
			/* hit in left arrow: small shift down in time (may repeat) */
			sim_waveformwindow = w;
			sim_waveformsliderpart = 0;
			trackcursor(0, us_nullup, us_nullvoid, sim_window_horizontalslider, us_nullchar,
				us_nullvoid, TRACKNORMAL);
			return;
		}
		if (inx < w->thumblx)
		{
			/* hit to left of thumb: large shift down in time (may repeat) */
			sim_waveformwindow = w;
			sim_waveformsliderpart = 1;
			trackcursor(0, us_nullup, us_nullvoid, sim_window_horizontalslider, us_nullchar,
				us_nullvoid, TRACKNORMAL);
			return;
		}
		if (inx <= w->thumbhx)
		{
			/* hit on thumb: track thumb motion */
			us_hthumbbegin(inx, w, w->usely, w->uselx, w->usehx);
			trackcursor(0, us_nullup, us_nullvoid, us_hthumbdown, us_nullchar,
				us_shthumbdone, TRACKNORMAL); 
			newpos = us_shthumbend();
			if (newpos == sim_window_mintime) return;
			sim_window_maxtime += newpos - sim_window_mintime;
			sim_window_mintime = newpos;
			simals_initialize_simulator(1);
			return;
		}
		if (inx < w->usehx - DISPLAYSLIDERSIZE)
		{
			/* hit to right of thumb: large shift up in time (may repeat) */
			sim_waveformwindow = w;
			sim_waveformsliderpart = 3;
			trackcursor(0, us_nullup, us_nullvoid, sim_window_horizontalslider, us_nullchar,
				us_nullvoid, TRACKNORMAL);
			return;
		}

		/* hit in right arrow: small shift up in time (may repeat) */
		sim_waveformwindow = w;
		sim_waveformsliderpart = 4;
		trackcursor(0, us_nullup, us_nullvoid, sim_window_horizontalslider, us_nullchar,
			us_nullvoid, TRACKNORMAL);
		return;
	}

	sim_window_setviewport(w);
	x = inx;   y = iny;
	sim_window_scaletowindow(w, &x, &y);

	/* see if a signal name was hit */
	if (x < 114)
	{
		temp_trace_spacing = 540 / sim_window_vislines;
		ind = sim_window_vislines - 1 - (y - 31) / temp_trace_spacing;
		trprev = NODISCHANNEL;
		for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
		{
			if ((tr->flags&TRACENOTDRAWN) == 0)
			{
				if (tr->position == ind + sim_window_topline)
				{
					ypos = (sim_window_vislines - 1 - ind) * temp_trace_spacing + 31 + tr->nameoff;
					if (y >= ypos && y <= ypos + sim_window_txthei) break;
				}
			}
			trprev = tr;
		}
		if (tr == NODISCHANNEL)
		{
			/* not over any signal: drag out selection of signal names */
			sim_window_initialx = x;
			sim_window_initialy = y;
			sim_window_currentdragy = y;
			trackcursor(0, us_nullup, us_nullvoid, sim_window_seleachdown,
				us_nullchar, us_nullvoid, TRACKDRAGGING);
			sim_window_drawtraceselect(w, sim_window_currentdragy, sim_window_initialy, 0);
			return;
		}

		/* see if it is a double-click on a bus name */
		if (doublebutton(button) != 0)
		{
			if ((tr->flags&TRACETYPE) == TRACEISBUS)
			{
				/* closed bus: open it up */
				for(otr = sim_window_plot; otr != NODISCHANNEL; otr = otr->nextdischannel)
					if (otr->buschannel == tr) otr->flags &= ~TRACENOTDRAWN;
				tr->flags = (tr->flags & ~TRACETYPE) | TRACEISOPENBUS;

				/* renumber lines in the display */
				for(i=0, otr = sim_window_plot; otr != NODISCHANNEL; otr = otr->nextdischannel)
					if ((otr->flags&TRACENOTDRAWN) == 0) otr->position = i++;
				sim_window_lines = i;

				/* adjust the number of visible lines */
				sim_window_setviewport(w);

				sim_window_redraw();
				return;
			}
			if ((tr->flags&TRACETYPE) == TRACEISOPENBUS)
			{
				/* opened bus: close it up */
				for(otr = sim_window_plot; otr != NODISCHANNEL; otr = otr->nextdischannel)
					if (otr->buschannel == tr) otr->flags |= TRACENOTDRAWN;
				tr->flags = (tr->flags & ~TRACETYPE) | TRACEISBUS;

				/* renumber lines in the display */
				for(i=0, otr = sim_window_plot; otr != NODISCHANNEL; otr = otr->nextdischannel)
					if ((otr->flags&TRACENOTDRAWN) == 0) otr->position = i++;
				sim_window_lines = i;

				sim_window_redraw();
				return;
			}
			return;
		}

		/* see if the signal shares a line with any others */
		for(otr = sim_window_plot; otr != NODISCHANNEL; otr = otr->nextdischannel)
		{
			if ((otr->flags&TRACENOTDRAWN) != 0) continue;
			if (otr != tr && otr->position == tr->position) break;
		}
		if (otr == NODISCHANNEL)
		{
			/* not a shared signal: handle dragging of the name */
			sim_window_tracedragyoff = 0;
			sim_window_initialy = y;
			sim_window_tracedragtr = tr;
			sim_window_drawtracedrag(w, 1);
			trackcursor(0, us_nullup, us_nullvoid, sim_window_eachdown,
				us_nullchar, us_nullvoid, TRACKDRAGGING);
			sim_window_drawtracedrag(w, 0);

			/* see if the signal name was dragged to a new location */
			tracelinesmoved = sim_window_tracedragyoff / temp_trace_spacing;
			if (tracelinesmoved != 0)
			{
				/* erase highlighting */
				sim_window_cleartracehighlight();

				/* find which signal name it was dragged to */
				otrprev = NODISCHANNEL;
				for(otr = sim_window_plot; otr != NODISCHANNEL; otr = otr->nextdischannel)
				{
					if ((otr->flags&TRACENOTDRAWN) == 0)
					{
						if (otr->position == tr->position - tracelinesmoved) break;
					}
					if (otr != tr) otrprev = otr;
				}
				if (otr != NODISCHANNEL)
				{
					/* pull out the selected signal */
					sim_window_removetrace(tr);

					/* insert signal in new location */
					if (tr->position < otr->position)
					{
						sim_window_addtrace(tr, otr);
					} else
					{
						sim_window_addtrace(tr, otrprev);
					}

					/* move any bus signals inside of this */
					buschain = endbuschain = NODISCHANNEL;
					for(otr = sim_window_plot; otr != NODISCHANNEL; otr = nexttr)
					{
						nexttr = otr->nextdischannel;
						if (otr->buschannel == tr)
						{
							sim_window_removetrace(otr);
							if (endbuschain == NODISCHANNEL) endbuschain = buschain = otr; else
							{
								endbuschain->nextdischannel = otr;
								endbuschain = otr;
							}
							otr->nextdischannel = NODISCHANNEL;
						}
					}
					addloc = tr;
					for(otr = buschain; otr != NODISCHANNEL; otr = nexttr)
					{
						nexttr = otr->nextdischannel;
						sim_window_addtrace(otr, addloc);
						addloc = otr;
					}

					/* renumber lines in the display */
					for(i=0, otr = sim_window_plot; otr != NODISCHANNEL; otr = otr->nextdischannel)
						if ((otr->flags&TRACENOTDRAWN) == 0) otr->position = i++;

					/* redisplay window */
					sim_window_redraw();

					/* save the new configuration */
					sim_window_savesignalorder();
				}
				sim_window_addhighlighttrace((INTBIG)tr);
				par[0] = "highlight";
				tellaid(net_aid, 1, par);
				return;
			}
		}

		/* highlight the selected signal */
		if (shiftbutton(button) == 0)
		{
			/* shift not held: simply highlight this signal */
			sim_window_cleartracehighlight();
			sim_window_addhighlighttrace((INTBIG)tr);
		} else
		{
			/* complement state of this signal */
			for(j=0; j<sim_window_highlightedtracecount; j++)
				if (tr == sim_window_highlightedtraces[j]) break;
			if (j < sim_window_highlightedtracecount)
			{
				/* signal already highlighted: unhighlight it */
				sim_window_deletehighlighttrace((INTBIG)tr);
			} else
			{
				/* signal new: add to highlight */
				sim_window_addhighlighttrace((INTBIG)tr);
			}

		}
		par[0] = "highlight";
		tellaid(net_aid, 1, par);
		return;
	}

	/* see if a cursor was hit */
	xpos = sim_window_timetoxpos(sim_window_maincursor);
	if (abs(x-xpos) < 10)
	{
		/* main cursor hit */
		trackcursor(0, us_nullup, us_nullvoid, sim_window_eachwdown,
			us_nullchar, us_nullvoid, TRACKDRAGGING);
		return;
	}
	xpos = sim_window_timetoxpos(sim_window_extensioncursor);
	if (abs(x-xpos) < 10)
	{
		/* extension cursor hit */
		trackcursor(0, us_nullup, us_nullvoid, sim_window_eachbdown,
			us_nullchar, us_nullvoid, TRACKDRAGGING);
		return;
	}
}

/*
 * This procedure redraws the display, allowing for resize.
 */
void sim_window_redisphandler(WINDOWPART *w)
{
	INTSML lx, hx, ly, hy;
	float range;
	REGISTER INTBIG i, thumbtop, thumbarea, thumbsize;

	sim_window_setviewport(w);

	/* clear the window */
	sim_window_desc.bits = LAYERA;
	sim_window_desc.col = ALLOFF;
	w->uselx -= DISPLAYSLIDERSIZE;
	w->usely -= DISPLAYSLIDERSIZE;
	lx = w->uselx;   hx = w->usehx;
	ly = w->usely;   hy = w->usehy;
	screendrawbox(w, lx, hx, ly, hy, &sim_window_desc);

	/* draw red outline */
	if ((w->state&WINDOWSIMULATING) != 0)
	{
		sim_window_desc.col = RED;
		screendrawbox(w, (INTSML)(lx-SIMULATINGBORDERSIZE), (INTSML)(hx+SIMULATINGBORDERSIZE),
			(INTSML)(ly-SIMULATINGBORDERSIZE), (INTSML)(ly-1), &sim_window_desc);
		screendrawbox(w, (INTSML)(lx-SIMULATINGBORDERSIZE), (INTSML)(hx+SIMULATINGBORDERSIZE),
			(INTSML)(hy+1), (INTSML)(hy+SIMULATINGBORDERSIZE), &sim_window_desc);
		screendrawbox(w, (INTSML)(lx-SIMULATINGBORDERSIZE), (INTSML)(lx-1), ly, hy, &sim_window_desc);
		screendrawbox(w, (INTSML)(hx+1), (INTSML)(hx+SIMULATINGBORDERSIZE), ly, hy, &sim_window_desc);
	}

	/* draw vertical slider on the left */
	us_drawverticalslider(w, (INTSML)(lx-2), (INTSML)(ly+DISPLAYSLIDERSIZE), hy, 1);
	thumbsize = sim_window_vislines * 100 / sim_window_lines;
	if (thumbsize >= 100) thumbsize = 100;
	if (thumbsize == 100 && sim_window_topline == 0)
	{
		w->thumbly = 0;   w->thumbhy = -1;
	} else
	{
		thumbtop = sim_window_topline * 100 / sim_window_lines;
		thumbarea = w->usehy - w->usely - DISPLAYSLIDERSIZE*2;

		w->thumbhy = w->usehy - DISPLAYSLIDERSIZE - thumbarea * thumbtop / 100;
		w->thumbly = w->thumbhy - thumbarea * thumbsize / 100;
		if (w->thumbhy > w->usehy-DISPLAYSLIDERSIZE-2) w->thumbhy = w->usehy-DISPLAYSLIDERSIZE-2;
		if (w->thumbly < w->usely+DISPLAYSLIDERSIZE*2+2) w->thumbly = w->usely+DISPLAYSLIDERSIZE*2+2;
	}
	us_drawverticalsliderthumb(w, lx-2, ly+DISPLAYSLIDERSIZE, hy, w->thumbly, w->thumbhy);

	/* draw horizontal slider on the bottom */
	us_drawhorizontalslider(w, (INTSML)(ly+DISPLAYSLIDERSIZE), (INTSML)(lx+DISPLAYSLIDERSIZE), hx);
	thumbarea = w->usehx - w->uselx - DISPLAYSLIDERSIZE*2;
	if (sim_window_mintime <= 0.0)
	{
		w->thumblx = w->uselx+DISPLAYSLIDERSIZE*2+2;
		w->thumbhx = w->thumblx + thumbarea/2;
	} else
	{
		range = sim_window_maxtime * 2.0f;
		w->thumblx = (INTSML)(w->uselx+DISPLAYSLIDERSIZE*2+2 + (INTBIG)(sim_window_mintime/range*thumbarea));
		w->thumbhx = (INTSML)(w->uselx+DISPLAYSLIDERSIZE*2+2 + (INTBIG)(sim_window_maxtime/range*thumbarea));
	}
	if (w->thumbhx > w->usehx-DISPLAYSLIDERSIZE-2) w->thumbhx = w->usehx-DISPLAYSLIDERSIZE-2;
	if (w->thumblx < w->uselx+DISPLAYSLIDERSIZE*2+2) w->thumblx = w->uselx+DISPLAYSLIDERSIZE*2+2;
	us_drawhorizontalsliderthumb(w, ly+DISPLAYSLIDERSIZE, lx+DISPLAYSLIDERSIZE, hx, w->thumblx, w->thumbhx);

	us_drawslidercorner(w, lx, lx+DISPLAYSLIDERSIZE-1, ly, ly+DISPLAYSLIDERSIZE-1);

	/* reset bounds */
	w->uselx += DISPLAYSLIDERSIZE;
	w->usely += DISPLAYSLIDERSIZE;

	/* write banner at top of window */
	screensettextsize(w, 12);
	sim_window_desc.col = el_colmentxt;
	(void)initinfstr();
	if ((w->state&WINDOWSIMULATING) == 0)
		(void)addstringtoinfstr(_("Inactive Simulation Window, ? for help")); else
			(void)addstringtoinfstr(_("Simulation Window, ? for help"));
	lx = w->uselx;
	screendrawtext(w, (INTSML)(lx+5), (INTSML)(hy-18), returninfstr(), &sim_window_desc);
	screeninvertbox(w, lx, hx, (INTSML)(hy-20), hy);

	sim_window_drawtimescale(w);
	sim_window_drawgraph(w);
	sim_window_drawcursors(w);
	for(i=0; i<sim_window_highlightedtracecount; i++)
		sim_window_plottrace(sim_window_highlightedtraces[i], w, 1);
}

/****************************** SUPPORT FUNCTIONS ******************************/

/*
 * Routine to convert from simulation time to the proper X position in the waveform window.
 */
INTBIG sim_window_timetoxpos(float time)
{
	float result;

	result = 653.0f * (time - sim_window_mintime) / (sim_window_maxtime - sim_window_mintime) + 114.0f;
	return(roundfloat(result));
}

/*
 * Routine to convert from the X position in the waveform window to simulation time.
 */
float sim_window_xpostotime(INTBIG xpos)
{
	float result;

	result = (xpos-114.0f) * (sim_window_maxtime-sim_window_mintime) / 653.0f + sim_window_mintime;
	return(result);
}

/*
 * This procedure prints out the time scale on the graphics terminal.
 */
void sim_window_drawtimescale(WINDOWPART *wavewin)
{
	INTBIG i, x_pos;
	float time;
	char s2[20];

	sim_window_setmask(LAYERA);
	sim_window_setcolor(MENTXT);
	for(i=1; i<5; i++)
	{
		time = i * 0.2f * (sim_window_maxtime - sim_window_mintime) + sim_window_mintime;
		x_pos = sim_window_timetoxpos(time);
		sim_windowconvertengineeringnotation(time, s2);
		sim_window_moveto(x_pos, 0);
		sim_window_drawcstring(wavewin, s2);
	}
}

void sim_window_writetracename(WINDOWPART *wavewin, DISCHANNEL *tr)
{
	INTBIG y_pos, temp_trace_spacing, line, px, py, maxheight, maxwidth, len, save, size;
	INTSML wid, hei;

	/* compute positions */
	temp_trace_spacing = 540 / sim_window_vislines;
	line = tr->position - sim_window_topline;
	if (line < 0 || line >= sim_window_vislines) return;
	y_pos = (sim_window_vislines - 1 - line) * temp_trace_spacing + 31;

	/* draw the vertical line between the label and the trace */
	sim_window_setcolor(MENGLY);
	sim_window_moveto(113, y_pos);
	sim_window_drawto(wavewin, 113, y_pos+temp_trace_spacing-10);

	/* set the text color */
	if ((tr->flags&TRACETYPE) == TRACEISANALOG) sim_window_setcolor(tr->color); else
		sim_window_setcolor(MENTXT);

	/* write the text */
	if (tr->buschannel == NODISCHANNEL) px = 1; else px = 10;
	py = y_pos + tr->nameoff;
	maxheight = applyyscale(wavewin, temp_trace_spacing);
	maxwidth = applyxscale(wavewin, 112-px);
	sim_window_mapcoord(wavewin, &px, &py);
	len = strlen(tr->name);
	for(size = sim_window_textsize; size >= 4; size--)
	{
		screensettextsize(wavewin, size);
		screengettextsize(wavewin, tr->name, &wid, &hei);
		if (hei > maxheight && size > 4) continue;
		sim_window_txthei = hei * (wavewin->screenhy - wavewin->screenly) /
			(wavewin->usehy - wavewin->usely - 20);
		while (len > 1 && wid > maxwidth)
		{
			len--;
			save = tr->name[len];   tr->name[len] = 0;
			screengettextsize(wavewin, tr->name, &wid, &hei);
			tr->name[len] = (char)save;
		}
		save = tr->name[len];   tr->name[len] = 0;
		screendrawtext(wavewin, (INTSML)px, (INTSML)py, tr->name, &sim_window_desc);
		tr->name[len] = (char)save;
		break;
	}
}

/*
 * This procedure draws the cursors on the timing diagram.  The other color
 * planes are write protected so that the cursor doesn't alter any of the
 * timing diagram information.
 */
void sim_window_drawcursors(WINDOWPART *wavewin)
{
	INTBIG x_pos;
	char s2[40];
	float v;

	/* erase all cursors */
	sim_window_setcolor(ALLOFF);
	sim_window_setmask(LAYERH);
	sim_window_drawbox(wavewin, 114, 0, 767, 570);
	sim_window_setcolor(HIGHLIT);

	/* draw main cursor */
	x_pos = sim_window_timetoxpos(sim_window_maincursor);
	if (x_pos >= 114 && x_pos <= 767)
	{
		sim_window_moveto(x_pos, 31);
		sim_window_drawto(wavewin, x_pos, 570);
	}

	/* draw extension cursor */
	x_pos = sim_window_timetoxpos(sim_window_extensioncursor);
	if (x_pos >= 114 && x_pos <= 767)
	{
		sim_window_moveto(x_pos, 31);
		sim_window_drawto(wavewin, x_pos, 570);
		sim_window_moveto(x_pos-5, 558);
		sim_window_drawto(wavewin, x_pos+5, 568);
		sim_window_moveto(x_pos+5, 558);
		sim_window_drawto(wavewin, x_pos-5, 568);
	}

	/* draw cursor locations textually */
	sim_window_setmask(LAYERA);
	sim_window_setcolor(ALLOFF);
	sim_window_drawbox(wavewin, 114, 571, 767, 600);		/* cursor text area */
	sim_window_setcolor(MENTXT);
	sim_window_moveto(114, 600);
	strcpy(s2, "Main: ");
	sim_windowconvertengineeringnotation(sim_window_maincursor, &s2[6]);
	sim_window_drawulstring(wavewin, s2);
	sim_window_moveto(333, 600);
	strcpy(s2, "Ext: ");
	sim_windowconvertengineeringnotation(sim_window_extensioncursor, &s2[5]);
	sim_window_drawulstring(wavewin, s2);
	sim_window_moveto(552, 600);
	strcpy(s2, "Delta: ");
	v = sim_window_extensioncursor - sim_window_maincursor;
	if (v < 0.0) v = -v;
	sim_windowconvertengineeringnotation(v, &s2[7]);
	sim_window_drawulstring(wavewin, s2);
}

void sim_window_drawgraph(WINDOWPART *wavewin)
{
	INTBIG y_min, y_max, j, maxoff,
		slot, temp_trace_spacing, range, hasana;
	static INTBIG sigtotal = 0;
	static INTBIG *sigvalues;
	INTSML x, y;
	char s2[15];
	REGISTER DISCHANNEL *tr, *otr;

	/* determine height of text in the window */
	screensettextsize(wavewin, sim_window_textsize);
	screengettextsize(wavewin, "Xy", &x, &y);
	sim_window_txthei = y * (wavewin->screenhy - wavewin->screenly) /
		(wavewin->usehy - wavewin->usely - 20);

	/* set name offsets */
	for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel) tr->nameoff = -1;
	temp_trace_spacing = 540 / sim_window_vislines;
	for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
	{
		/* determine position of trace label on the line */
		if ((tr->flags&TRACENOTDRAWN) != 0) continue;
		if (tr->nameoff < 0)
		{
			hasana = 0;
			maxoff = 0;
			for(otr = sim_window_plot; otr != NODISCHANNEL; otr = otr->nextdischannel)
			{
				if ((otr->flags&TRACENOTDRAWN) != 0) continue;
				if (otr->position != tr->position) continue;
				maxoff++;
				if ((otr->flags&TRACETYPE) == TRACEISANALOG) hasana++;
			}
			if (maxoff == 1)
			{
				/* only one trace: center it */
				tr->nameoff = (temp_trace_spacing - 10 - sim_window_txthei) / 2;
				if (tr->nameoff < 0) tr->nameoff = 0;
			} else
			{
				/* multiple traces: compute them */
				range = (temp_trace_spacing - 10 - sim_window_txthei*3) / (maxoff-1);
				if (range <= 0) range = 1;
				slot = 0;
				for(otr = sim_window_plot; otr != NODISCHANNEL; otr = otr->nextdischannel)
				{
					if (otr->position != tr->position) continue;
					otr->nameoff = slot * range + sim_window_txthei;
					slot++;
				}
			}

			/* if there are analog traces on this line, show height values */
			if (hasana != 0)
			{
				y_min = (sim_window_vislines - 1 - (tr->position - sim_window_topline)) * temp_trace_spacing + 31;
				y_max = y_min + temp_trace_spacing - 10;
				sim_window_setmask(LAYERA);
				sim_window_setcolor(MENTXT);
				sim_window_moveto(100, y_min);
				sim_window_drawto(wavewin, 113, y_min);
				sim_window_moveto(112, y_min);
				(void)sprintf(s2, "%g", sim_window_anadislow);
				sim_window_drawrstring(wavewin, s2);

				sim_window_moveto(100, y_max);
				sim_window_drawto(wavewin, 113, y_max);
				sim_window_moveto(112, y_max-sim_window_txthei);
				(void)sprintf(s2, "%g", sim_window_anadishigh);
				sim_window_drawrstring(wavewin, s2);
			}
		}
	}

	/* set colors for analog traces */
	j = 0;
	for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
	{
		if ((tr->flags&TRACENOTDRAWN) != 0) continue;

		/* handle analog trace color */
		if ((tr->flags&TRACETYPE) == TRACEISANALOG)
		{
			tr->color = sim_window_anacolor[j];
			j++;
			if (sim_window_anacolor[j] == 0) j = 0;
		}
	}

	/* plot it all */
	for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
	{
		if ((tr->flags&TRACENOTDRAWN) != 0) continue;
		if (stopping(STOPREASONDISPLAY)) break;
		sim_window_plottrace(tr, wavewin, 0);
	}
}

/*
 * Routine to draw a trace "tr" in waveform window "wavewin".  "style" is:
 *  0   draw normal trace
 *  1   highlight trace
 *  2   unhighlight trace
 */
void sim_window_plottrace(DISCHANNEL *tr, WINDOWPART *wavewin, INTBIG style)
{
	INTBIG x_min, x_max, y_min, y_max, j, state, laststate, first, windowstate,
		slot, temp_trace_spacing, fx, fy, tx, ty, color, lx, hx, ly, hy, strength,
		cfx, cfy, ctx, cty, sigcount, newtotal, undefined, val, i, flags;
	static INTBIG sigtotal = 0;
	static INTBIG *sigvalues;
	char s2[50];
	REGISTER DISCHANNEL *bustr;
	REGISTER VARIABLE *var;
	float time, position, nexttime;

	/* get coordinates */
	temp_trace_spacing = 540 / sim_window_vislines;
	slot = tr->position - sim_window_topline;
	if (slot < 0 || slot >= sim_window_vislines) return;
	y_min = (sim_window_vislines - 1 - slot) * temp_trace_spacing + 31;
	y_max = y_min + temp_trace_spacing - 10;

	/* handle name field on the left */
	if (style == 0)
	{
		/* draw normal trace name */
		sim_window_writetracename(wavewin, tr);
	} else
	{
		/* highlight/unhighlight trace name */
		ly = y_min + tr->nameoff+1;   hy = ly + sim_window_txthei;
		lx = 0;                       hx = 112;
		sim_window_setmask(LAYERH);
		if (style == 1) color = HIGHLIT; else color = 0;
		sim_window_setcolor(color);
		sim_window_moveto(lx, ly);
		sim_window_drawto(wavewin, hx, ly);
		sim_window_drawto(wavewin, hx, hy);
		sim_window_drawto(wavewin, lx, hy);
		sim_window_drawto(wavewin, lx, ly);
	}

	/* handle busses */
	flags = tr->flags & TRACETYPE;
	if (flags == TRACEISBUS || flags == TRACEISOPENBUS)
	{
		/* determine base of bus */
		var = getvalkey((INTBIG)sim_aid, VAID, VINTEGER, sim_window_state);
		if (var == NOVARIABLE) windowstate = 0; else windowstate = var->addr;

		/* find starting time and enumerate signals */
		sigcount = 0;
		for(bustr = sim_window_plot; bustr != NODISCHANNEL; bustr = bustr->nextdischannel)
		{
			if (bustr->buschannel != tr) continue;
			if (sigcount == 0 || bustr->timearray[0] < time)
				time = bustr->timearray[0];
			bustr->busindex = sigcount++;
		}

		/* make sure there is room for all of the signal values */
		if (sigcount > sigtotal)
		{
			newtotal = sigtotal * 2;
			if (newtotal == 0) newtotal = 8;
			if (newtotal < sigcount) newtotal = sigcount;
			if (sigtotal > 0) efree((char *)sigvalues);
			sigtotal = 0;
			sigvalues = (INTBIG *)emalloc(newtotal * SIZEOFINTBIG, sim_aid->cluster);
			if (sigvalues == 0) return;
			sigtotal = newtotal;
		}

		/* draw the waveform */
		x_min = 114;
		for(;;)
		{
			/* compute signal value at time "time" */
			undefined = 0;
			for(bustr = sim_window_plot; bustr != NODISCHANNEL; bustr = bustr->nextdischannel)
			{
				if (bustr->buschannel != tr) continue;
				for(j=0; j<bustr->numsteps; j++)
					if (time < bustr->timearray[j]) break;
				if (j > 0) j--;
				sigvalues[bustr->busindex] = bustr->statearray[j] >> 8;
				if (sigvalues[bustr->busindex] != LOGIC_LOW && sigvalues[bustr->busindex] != LOGIC_HIGH)
				{
					undefined = 1;
					break;
				}
			}

			/* compute the value string */
			if (undefined != 0) strcpy(s2, "XX"); else
			{
				val = 0;
				for(i=0; i<sigcount; i++)
				{
					val <<= 1;
					if (sigvalues[i] == LOGIC_HIGH) val |= 1;
				}
				switch (windowstate&BUSBASEBITS)
				{
					case BUSBASE2:
						strcpy(s2, "0b");
						for(i=0; i<sigcount; i++)
							if (sigvalues[i] == LOGIC_HIGH)
								strcat(s2, "1"); else
									strcat(s2, "0");
						break;
					case BUSBASE8:  sprintf(s2, "0%o", val);    break;
					case BUSBASE10: sprintf(s2, "%ld", val);    break;
					case BUSBASE16: sprintf(s2, "0x%X", val);   break;
				}
			}

			/* draw the value */
			x_max = sim_window_timetoxpos(time);
			if (x_max < 114) x_max = 114;
			if (x_max > 767) break;
#define BUSANGLE 4
			if (x_max > x_min)
			{
				if (style == 0)
					sim_window_setcolor(sim_colorstrengthpower);
				sim_window_moveto(x_min, (y_min+y_max)/2);
				sim_window_drawto(wavewin, x_min+BUSANGLE, y_min);
				sim_window_drawto(wavewin, x_max-BUSANGLE, y_min);
				sim_window_drawto(wavewin, x_max, (y_min+y_max)/2);
				sim_window_drawto(wavewin, x_max-BUSANGLE, y_max);
				sim_window_drawto(wavewin, x_min+BUSANGLE, y_max);
				sim_window_drawto(wavewin, x_min, (y_min+y_max)/2);
				if (style == 0)
				{
					sim_window_moveto(x_max+BUSANGLE/2, (y_min+y_max)/2);
					sim_window_drawhcstring(wavewin, s2);
				}
				x_min = x_max;
			}

			/* advance to the next time */
			nexttime = time;
			for(bustr = sim_window_plot; bustr != NODISCHANNEL; bustr = bustr->nextdischannel)
			{
				if (bustr->buschannel != tr) continue;
				for(j=0; j<bustr->numsteps; j++)
					if (bustr->timearray[j] > time) break;
				if (j < bustr->numsteps)
				{
					if (nexttime == time || bustr->timearray[j] < nexttime)
						nexttime = bustr->timearray[j];
				}
			}
			if (nexttime == time) break;
			time = nexttime;
		}
		if (x_max < 767)
		{
			x_min = x_max;   x_max = 767;
			if (style == 0)
				sim_window_setcolor(sim_colorstrengthpower);
			sim_window_moveto(x_min, (y_min+y_max)/2);
			sim_window_drawto(wavewin, x_min+BUSANGLE, y_min);
			sim_window_drawto(wavewin, x_max-BUSANGLE, y_min);
			sim_window_drawto(wavewin, x_max, (y_min+y_max)/2);
			sim_window_drawto(wavewin, x_max-BUSANGLE, y_max);
			sim_window_drawto(wavewin, x_min+BUSANGLE, y_max);
			sim_window_drawto(wavewin, x_min, (y_min+y_max)/2);
		}
		if (style != 0) sim_window_setmask(LAYERA);
		return;
	}

	/* handle digital waveform */
	if (flags == TRACEISDIGITAL)
	{
		first = 1;
		for(j=0; j<tr->numsteps; j++)
		{
			time = tr->timearray[j];
			if (j == tr->numsteps-1) x_max = 767; else
			{
				x_max = sim_window_timetoxpos(tr->timearray[j+1]);
				if (x_max < 114) continue;
			}
			x_min = sim_window_timetoxpos(time);
			if (x_min < 114) x_min = 114;
			if (x_min > 767) continue;

			state = tr->statearray[j] >> 8;
			if (style == 0)
			{
				strength = tr->statearray[j] & 0377;
				if (strength == 0) color = sim_colorstrengthoff; else
					if (strength <= NODE_STRENGTH) color = sim_colorstrengthnode; else
						if (strength <= GATE_STRENGTH) color = sim_colorstrengthgate; else
							color = sim_colorstrengthpower;
				sim_window_setcolor(color);
			}
			if (first == 0 && state != laststate)
			{
				sim_window_moveto(x_min, y_min);
				sim_window_drawto(wavewin, x_min, y_max);
			}
			first = 0;
			switch (state)
			{
				case LOGIC_LOW:
					sim_window_moveto(x_min, y_min);
					sim_window_drawto(wavewin, x_max, y_min);
					break;
				case LOGIC_X:
					sim_window_drawbox(wavewin, x_min, y_min, x_max, y_max);
					break;
				case LOGIC_HIGH:
					sim_window_moveto(x_min, y_max);
					sim_window_drawto(wavewin, x_max, y_max);
					break;
				default:
					sim_window_drawbox(wavewin, x_min, y_min, x_max, y_max);
					if (style == 0)
					{
						if ((x_max - x_min) <= 1) break;
						sim_window_setcolor(ALLOFF);
						sim_window_drawbox(wavewin, x_min+1, y_min+1, x_max-1, y_max-1);
						(void)sprintf(s2, "0x%X", state);
						sim_window_moveto((x_min + x_max) / 2, y_min+3);
						sim_window_setcolor(color);
						sim_window_drawcstring(wavewin, s2);
					}
					break;
			}
			laststate = state;
		}
		if (style != 0) sim_window_setmask(LAYERA);
		return;
	}

	/* handle analog waveform */
	first = 1;
	for(j=0; j<tr->numsteps; j++)
	{
		time = tr->timearray[j];
		tx = sim_window_timetoxpos(time);
		position = (float)(y_max - y_min);
		position *= (tr->valuearray[j] * sim_window_anarange - sim_window_anadislow) /
			sim_window_anadisrange;
		ty = (INTBIG)position + y_min;
		if (j != 0 && tx >= 114)
		{
			if (tx > 767)
			{
				/* "to" data point off the screen: interpolate */
				ty = (ty-fy) * (767-fx) / (tx-fx) + fy;
				tx = 767;
			}

			if (first != 0 && fx < 114)
			{
				/* "from" data point off the screen: interpolate */
				fy = (ty-fy) * (114-fx) / (tx-fx) + fy;
				fx = 114;
			}
			first = 0;
			if (style == 0) sim_window_setcolor(tr->color);
			if (fy >= y_min && fy <= y_max &&
				ty >= y_min && ty <= y_max)
			{
				sim_window_moveto(fx, fy);
				sim_window_drawto(wavewin, tx, ty);
			} else
			{
				cfx = fx;   cfy = fy;
				ctx = tx;   cty = ty;
				if (clipline(&cfx, &cfy, &ctx, &cty, 114, 767, y_min, y_max) == 0)
				{
					sim_window_moveto(cfx, cfy);
					sim_window_drawto(wavewin, ctx, cty);
				}
			}
		}
		fx = tx;   fy = ty;
		if (fx >= 767) break;
	}
	if (style != 0) sim_window_setmask(LAYERA);
}

/*
 * routine to snapshot the simulation window in a facet with artwork primitives
 */
void sim_window_savegraph(void)
{
	INTBIG x_min, x_max, y_min, y_max, j, maxoff, state, laststate, color, first,
		slot, temp_trace_spacing, range, hasana, lastcolor, tx1, ty1, tx2, ty2,
		fx, fy, tx, ty, cfx, cfy, ctx, cty, datapos, strength;
	INTSML x, y;
	char s2[15];
	REGISTER INTBIG *data, maxsteps;
	REGISTER NODEPROTO *np, *plotnp;
	REGISTER NODEINST *ni;
	REGISTER VARIABLE *var;
	REGISTER WINDOWPART *wavewin;
	REGISTER DISCHANNEL *tr, *otr;
	float time, position;

	/* get facet being saved */
	np = getcurfacet();
	if (np == NONODEPROTO) return;

	/* get waveform window */
	wavewin = sim_window_findwaveform();
	if (wavewin == NOWINDOWPART) return;

	/* determine height of text in the window */
	screensettextsize(wavewin, sim_window_textsize);
	screengettextsize(wavewin, "Xy", &x, &y);
	sim_window_txthei = y * (wavewin->screenhy - wavewin->screenly) /
		(wavewin->usehy - wavewin->usely - 20);

	/* create the plot facet */
	(void)initinfstr();
	(void)addstringtoinfstr(np->cell->cellname);
	(void)addstringtoinfstr("{sim}");
	plotnp = newnodeproto(returninfstr(), el_curlib);
	if (plotnp == NONODEPROTO) return;
	setactivity("Make Simulation Plot");

	/* determine maximum number of steps */
	maxsteps = 1;
	for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
	{
		if ((tr->flags&TRACENOTDRAWN) != 0) continue;
		if (tr->numsteps > maxsteps) maxsteps = tr->numsteps;
	}

	/* make space */
	data = (INTBIG *)emalloc(maxsteps * 4 * SIZEOFINTBIG, el_tempcluster);
	if (data == 0) return;

	/* compute transformation for points in the plot area */
	tx1 = 114;   ty1 = 31;    sim_window_mapcoord(wavewin, &tx1, &ty1);
	tx2 = 767;   ty2 = 570;   sim_window_mapcoord(wavewin, &tx2, &ty2);
	sim_window_offx = (tx2 - tx1) / 2;   sim_window_basex = tx1;
	sim_window_offy = (ty2 - ty1) / 2;   sim_window_basey = ty1;

	/* write the header */
	tx1 = 383;   ty1 = 600;   sim_window_mapcoord(wavewin, &tx1, &ty1);
	ni = newnodeinst(gen_invispinprim, tx1, tx1, ty1, ty1, 0, 0, plotnp);
	if (ni == NONODEINST) { efree((char *)data);  return; }
	(void)initinfstr();
	(void)addstringtoinfstr("Simulation snapshot of facet ");
	(void)addstringtoinfstr(describenodeproto(np));
	var = setvalkey((INTBIG)ni, VNODEINST, art_messagekey, (INTBIG)returninfstr(), VSTRING|VDISPLAY);
	if (var != NOVARIABLE) var->textdescript = (TXT14P << VTSIZESH) | VTPOSUP;

	/* set name offsets */
	for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel) tr->nameoff = -1;
	temp_trace_spacing = 540 / sim_window_vislines;
	for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
	{
		if ((tr->flags&TRACENOTDRAWN) != 0) continue;

		/* determine position of trace label on the line */
		if (tr->nameoff < 0)
		{
			hasana = 0;
			maxoff = 0;
			for(otr = sim_window_plot; otr != NODISCHANNEL; otr = otr->nextdischannel)
			{
				if ((otr->flags&TRACENOTDRAWN) != 0) continue;
				if (otr->position != tr->position) continue;
				maxoff++;
				if ((otr->flags&TRACETYPE) == TRACEISANALOG) hasana++;
			}
			if (maxoff == 1)
			{
				/* only one trace: center it */
				tr->nameoff = (temp_trace_spacing - 10 - sim_window_txthei) / 2;
				if (tr->nameoff < 0) tr->nameoff = 0;
			} else
			{
				/* multiple traces: compute them */
				range = (temp_trace_spacing - 10 - sim_window_txthei*3) / (maxoff-1);
				if (range <= 0) range = 1;
				slot = 0;
				for(otr = sim_window_plot; otr != NODISCHANNEL; otr = otr->nextdischannel)
				{
					if ((otr->flags&TRACENOTDRAWN) != 0) continue;
					if (otr->position != tr->position) continue;
					otr->nameoff = slot * range + sim_window_txthei;
					slot++;
				}
			}

			/* if there are analog traces on this line, show height values */
			if (hasana != 0)
			{
				y_min = (sim_window_vislines - 1 - (tr->position - sim_window_topline)) * temp_trace_spacing + 31;
				y_max = y_min + temp_trace_spacing - 10;

				tx1 = 101;   ty1 = y_min;   sim_window_mapcoord(wavewin, &tx1, &ty1);
				tx2 = 113;   ty2 = y_min;   sim_window_mapcoord(wavewin, &tx2, &ty2);
				data[0] = -(tx2-tx1)/2;  data[1] = 0;
				data[2] = (tx2-tx1)/2;   data[3] = 0;
				ni = newnodeinst(art_openedpolygonprim, tx1,tx2, ty1,ty2, 0, 0, plotnp);
				if (ni == NONODEINST) { efree((char *)data);  return; }
				(void)setvalkey((INTBIG)ni, VNODEINST, el_trace, (INTBIG)data,
					VINTEGER|VISARRAY|(4<<VLENGTHSH));

				(void)sprintf(s2, "%g", sim_window_anadislow);
				tx1 = 112;   ty1 = y_min;   sim_window_mapcoord(wavewin, &tx1, &ty1);
				ni = newnodeinst(gen_invispinprim, tx1, tx1, ty1, ty1, 0, 0, plotnp);
				if (ni == NONODEINST) { efree((char *)data);  return; }
				var = setvalkey((INTBIG)ni, VNODEINST, art_messagekey, (INTBIG)s2, VSTRING|VDISPLAY);
				if (var != NOVARIABLE) var->textdescript = (TXT10P << VTSIZESH) | VTPOSUPLEFT;

				tx1 = 101;   ty1 = y_max;   sim_window_mapcoord(wavewin, &tx1, &ty1);
				tx2 = 113;   ty2 = y_max;   sim_window_mapcoord(wavewin, &tx2, &ty2);
				data[0] = -(tx2-tx1)/2;  data[1] = 0;
				data[2] = (tx2-tx1)/2;   data[3] = 0;
				ni = newnodeinst(art_openedpolygonprim, tx1,tx2, ty1,ty2, 0, 0, plotnp);
				if (ni == NONODEINST) { efree((char *)data);  return; }
				(void)setvalkey((INTBIG)ni, VNODEINST, el_trace, (INTBIG)data,
					VINTEGER|VISARRAY|(4<<VLENGTHSH));

				(void)sprintf(s2, "%g", sim_window_anadishigh);
				tx1 = 112;   ty1 = y_max;   sim_window_mapcoord(wavewin, &tx1, &ty1);
				ni = newnodeinst(gen_invispinprim, tx1, tx1, ty1, ty1, 0, 0, plotnp);
				if (ni == NONODEINST) { efree((char *)data);  return; }
				var = setvalkey((INTBIG)ni, VNODEINST, art_messagekey, (INTBIG)s2, VSTRING|VDISPLAY);
				if (var != NOVARIABLE) var->textdescript = (TXT10P << VTSIZESH) | VTPOSDOWNLEFT;
			}
		}
	}

	/* set colors for analog traces */
	j = 0;
	for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
	{
		/* handle analog trace color */
		if ((tr->flags&TRACENOTDRAWN) != 0) continue;
		if ((tr->flags&TRACETYPE) == TRACEISANALOG)
		{
			tr->color = sim_window_anacolor[j];
			j++;
			if (sim_window_anacolor[j] == 0) j = 0;
		}
	}

	/* plot time values */
	for(j=1; j<5; j++)
	{
		time = j * 0.2f * (sim_window_maxtime - sim_window_mintime) + sim_window_mintime;
		x = (INTSML)sim_window_timetoxpos(time);
		sim_windowconvertengineeringnotation(time, s2);
		tx1 = x;   ty1 = 0;   sim_window_mapcoord(wavewin, &tx1, &ty1);
		ni = newnodeinst(gen_invispinprim, tx1, tx1, ty1, ty1, 0, 0, plotnp);
		if (ni == NONODEINST) { efree((char *)data);  return; }
		var = setvalkey((INTBIG)ni, VNODEINST, art_messagekey, (INTBIG)s2, VSTRING|VDISPLAY);
		if (var != NOVARIABLE) var->textdescript = (TXT10P << VTSIZESH) | VTPOSUP;
	}

	/* plot it all */
	for(tr = sim_window_plot; tr != NODISCHANNEL; tr = tr->nextdischannel)
	{
		if ((tr->flags&TRACENOTDRAWN) != 0) continue;

		/* can't save bus traces yet */
		if ((tr->flags&TRACETYPE) == TRACEISBUS ||
			(tr->flags&TRACETYPE) == TRACEISOPENBUS) continue;

		if (stopping(STOPREASONDISPLAY)) break;

		/* compute positions */
		slot = tr->position - sim_window_topline;
		if (slot < 0 || slot >= sim_window_vislines) continue;
		y_min = (sim_window_vislines - 1 - slot) * temp_trace_spacing + 31;

		/* draw the vertical line between the label and the trace */
		tx1 = 113;   ty1 = y_min;   sim_window_mapcoord(wavewin, &tx1, &ty1);
		tx2 = 113;   ty2 = y_min+temp_trace_spacing-10;   sim_window_mapcoord(wavewin, &tx2, &ty2);
		data[0] = 0;   data[1] = -(ty2-ty1)/2;
		data[2] = 0;   data[3] = (ty2-ty1)/2;
		ni = newnodeinst(art_openedpolygonprim, tx1, tx2, ty1, ty2, 0, 0, plotnp);
		if (ni == NONODEINST) { efree((char *)data);  return; }
		(void)setvalkey((INTBIG)ni, VNODEINST, el_trace, (INTBIG)data,
			VINTEGER|VISARRAY|(4<<VLENGTHSH));

		/* write the text */
		(void)sprintf(s2, "%g", sim_window_anahigh);
		tx1 = 1;   ty1 = y_min + tr->nameoff;   sim_window_mapcoord(wavewin, &tx1, &ty1);
		ni = newnodeinst(gen_invispinprim, tx1, tx1, ty1, ty1, 0, 0, plotnp);
		if (ni == NONODEINST) { efree((char *)data);  return; }
		var = setvalkey((INTBIG)ni, VNODEINST, art_messagekey, (INTBIG)tr->name, VSTRING|VDISPLAY);
		if (var != NOVARIABLE) var->textdescript = (TXT10P << VTSIZESH) | VTPOSUPRIGHT;
		if ((tr->flags&TRACETYPE) == TRACEISANALOG)
		{
			tx1 = 0;     ty1 = y_min + tr->nameoff;
			sim_window_mapcoord(wavewin, &tx1, &ty1);
			tx2 = 100;   ty2 = y_min + tr->nameoff;
			sim_window_mapcoord(wavewin, &tx2, &ty2);
			data[0] = -(tx2-tx1)/2;  data[1] = 0;
			data[2] = (tx2-tx1)/2;   data[3] = 0;
			ni = newnodeinst(art_openedpolygonprim, tx1, tx2, ty1, ty2, 0, 0, plotnp);
			if (ni == NONODEINST) { efree((char *)data);  return; }
			(void)setvalkey((INTBIG)ni, VNODEINST, el_trace, (INTBIG)data,
				VINTEGER|VISARRAY|(4<<VLENGTHSH));
			(void)setvalkey((INTBIG)ni, VNODEINST, art_colorkey, tr->color, VINTEGER);
		}

		y_min = (sim_window_vislines - 1 - slot) * temp_trace_spacing + 31;
		y_max = y_min + temp_trace_spacing - 10;

		first = 1;
		datapos = 0;
		laststate = -1;
		for(j=0; j<tr->numsteps; j++)
		{
			time = tr->timearray[j];
			if ((tr->flags&TRACETYPE) == TRACEISDIGITAL)
			{
				/* digital waveform */
				if (j == tr->numsteps-1) x_max = 767; else
				{
					x_max = sim_window_timetoxpos(tr->timearray[j+1]);
					if (x_max < 114) continue;
				}
				x_min = sim_window_timetoxpos(time);
				if (x_min < 114) x_min = 114;
				if (x_min > 767) continue;

				strength = tr->statearray[j] & 0377;
				if (strength == 0) color = sim_colorstrengthoff; else
					if (strength <= NODE_STRENGTH) color = sim_colorstrengthnode; else
						if (strength <= GATE_STRENGTH) color = sim_colorstrengthgate; else
							color = sim_colorstrengthpower;
				if (datapos > 0 && color != lastcolor)
				{
					/* strength has changed, dump out data so far */
					tx1 = 114;   ty1 = 31;    sim_window_mapcoord(wavewin, &tx1, &ty1);
					tx2 = 767;   ty2 = 570;   sim_window_mapcoord(wavewin, &tx2, &ty2);
					ni = newnodeinst(art_openedpolygonprim, tx1, tx2, ty1, ty2, 0, 0, plotnp);
					if (ni == NONODEINST) { efree((char *)data);  return; }
					(void)setvalkey((INTBIG)ni, VNODEINST, el_trace, (INTBIG)data,
						VINTEGER|VISARRAY|(datapos<<VLENGTHSH));
					(void)setvalkey((INTBIG)ni, VNODEINST, art_colorkey, lastcolor, VINTEGER);
					datapos = 0;
				}
				lastcolor = color;

				state = tr->statearray[j] >> 8;
				switch (state)
				{
					case LOGIC_LOW:
						if (laststate == LOGIC_HIGH)
							sim_window_addsegment(wavewin, data, &datapos, x_min, y_max, x_min, y_min);
						sim_window_addsegment(wavewin, data, &datapos, x_min, y_min, x_max, y_min);
						break;
					case LOGIC_HIGH:
						if (laststate == LOGIC_LOW)
							sim_window_addsegment(wavewin, data, &datapos, x_min, y_min, x_min, y_max);
						sim_window_addsegment(wavewin, data, &datapos, x_min, y_max, x_max, y_max);
						break;
					default:
						if (datapos > 0)
						{
							/* doing unknown block, dump out any line data */
							tx1 = 114;   ty1 = 31;    sim_window_mapcoord(wavewin, &tx1, &ty1);
							tx2 = 767;   ty2 = 570;   sim_window_mapcoord(wavewin, &tx2, &ty2);
							ni = newnodeinst(art_openedpolygonprim, tx1, tx2, ty1, ty2, 0, 0, plotnp);
							if (ni == NONODEINST) { efree((char *)data);  return; }
							(void)setvalkey((INTBIG)ni, VNODEINST, el_trace, (INTBIG)data,
								VINTEGER|VISARRAY|(datapos<<VLENGTHSH));
							(void)setvalkey((INTBIG)ni, VNODEINST, art_colorkey, lastcolor,
								VINTEGER);
						}
						datapos = 0;
						tx1 = x_min;   ty1 = y_min;   sim_window_mapcoord(wavewin, &tx1, &ty1);
						tx2 = x_max;   ty2 = y_max;   sim_window_mapcoord(wavewin, &tx2, &ty2);
						ni = newnodeinst(art_boxprim, tx1, tx2, ty1, ty2, 0, 0, plotnp);
						if (ni == NONODEINST) { efree((char *)data);  return; }
						(void)setvalkey((INTBIG)ni, VNODEINST, art_colorkey, lastcolor, VINTEGER);
						break;
				}
				laststate = state;
				first = 0;
			} else
			{
				/* analog waveform */
				tx = sim_window_timetoxpos(time);
				position = (float)(y_max - y_min);
				position *= (tr->valuearray[j] * sim_window_anarange - sim_window_anadislow) /
					sim_window_anadisrange;
				ty = (INTBIG)position + y_min;
				if (j != 0 && tx >= 114)
				{
					if (tx > 767)
					{
						/* "to" data point off the screen: interpolate */
						ty = (ty-fy) * (767-fx) / (tx-fx) + fy;
						tx = 767;
					}

					if (first != 0 && fx < 114)
					{
						/* "from" data point off the screen: interpolate */
						fy = (ty-fy) * (114-fx) / (tx-fx) + fy;
						fx = 114;
					}
					first = 0;
					if (fy >= y_min && fy <= y_max &&
						ty >= y_min && ty <= y_max)
					{
						sim_window_addsegment(wavewin, data, &datapos, fx, fy, tx, ty);
					} else
					{
						cfx = fx;   cfy = fy;
						ctx = tx;   cty = ty;
						if (clipline(&cfx, &cfy, &ctx, &cty, 114, 767, y_min, y_max) == 0)
						{
							sim_window_addsegment(wavewin, data, &datapos, cfx, cfy, ctx, cty);
						}
					}
				}
				fx = tx;   fy = ty;
				if (fx >= 767) break;
			}
		}
		if (datapos > 0)
		{
			tx1 = 114;   ty1 = 31;    sim_window_mapcoord(wavewin, &tx1, &ty1);
			tx2 = 767;   ty2 = 570;   sim_window_mapcoord(wavewin, &tx2, &ty2);
			ni = newnodeinst(art_openedpolygonprim, tx1, tx2, ty1, ty2, 0, 0, plotnp);
			if (ni == NONODEINST) { efree((char *)data);  return; }
			(void)setvalkey((INTBIG)ni, VNODEINST, el_trace, (INTBIG)data,
				VINTEGER|VISARRAY|(datapos<<VLENGTHSH));
			if ((tr->flags&TRACETYPE) == TRACEISANALOG)
				(void)setvalkey((INTBIG)ni, VNODEINST, art_colorkey, tr->color, VINTEGER); else
					(void)setvalkey((INTBIG)ni, VNODEINST, art_colorkey, lastcolor, VINTEGER);
		}
	}

	/* ensure that the facet has the right size */
	(*el_curconstraint->solve)(plotnp);

	/* clean up */
	ttyputmsg("Created facet %s", describenodeproto(plotnp));
	efree((char *)data);
}

/*
 * Set the display color for strength "strength" to "color" (if "strength" is -1, show all colors).
 */
void sim_window_displaycolor(INTBIG strength, INTBIG color)
{
	char *colorname, *colorsymbol;

	if (strength == LOGIC_LOW) sim_colorlevellow = color; else
	if (strength == LOGIC_HIGH) sim_colorlevelhigh = color; else
	if (strength == LOGIC_X) sim_colorlevelundef = color; else
	if (strength == OFF_STRENGTH) sim_colorstrengthoff = color; else
	if (strength <= NODE_STRENGTH) sim_colorstrengthnode = color; else
	if (strength <= GATE_STRENGTH) sim_colorstrengthgate = color; else
	if (strength <= VDD_STRENGTH) sim_colorstrengthpower = color; else
	{
		if (ecolorname((INTSML)sim_colorstrengthoff, &colorname, &colorsymbol) != 0)
			colorname = "**UNKNOWN**";
		ttyputmsg("Off-strength signals are colored %s", colorname);
		if (ecolorname((INTSML)sim_colorstrengthnode, &colorname, &colorsymbol) != 0)
			colorname = "**UNKNOWN**";
		ttyputmsg("Node-strength signals are colored %s", colorname);
		if (ecolorname((INTSML)sim_colorstrengthgate, &colorname, &colorsymbol) != 0)
			colorname = "**UNKNOWN**";
		ttyputmsg("Gate-strength signals are colored %s", colorname);
		if (ecolorname((INTSML)sim_colorstrengthpower, &colorname, &colorsymbol) != 0)
			colorname = "**UNKNOWN**";
		ttyputmsg("Power-strength signals are colored %s", colorname);
		if (ecolorname((INTSML)sim_colorlevellow, &colorname, &colorsymbol) != 0)
			colorname = "**UNKNOWN**";
		ttyputmsg("Low-level signals are colored %s", colorname);
		if (ecolorname((INTSML)sim_colorlevelhigh, &colorname, &colorsymbol) != 0)
			colorname = "**UNKNOWN**";
		ttyputmsg("High-level signals are colored %s", colorname);
		if (ecolorname((INTSML)sim_colorlevelundef, &colorname, &colorsymbol) != 0)
			colorname = "**UNKNOWN**";
		ttyputmsg("Undefined-level signals are colored %s", colorname);
	}
}

void sim_window_addsegment(WINDOWPART *wavewin, INTBIG *data, INTBIG *datapos, INTBIG fx, INTBIG fy,
	INTBIG tx, INTBIG ty)
{
	REGISTER INTBIG truefx, truefy, pos;
	INTBIG x, y;

	x = fx;   y = fy;   sim_window_mapcoord(wavewin, &x, &y);
	truefx = (x-sim_window_basex) - sim_window_offx;
	truefy = (y-sim_window_basey) - sim_window_offy;

	pos = *datapos;
	if (pos < 2 || data[pos-2] != truefx || data[pos-1] != truefy)
	{
		data[pos++] = truefx;
		data[pos++] = truefy;
	}

	x = tx;   y = ty;   sim_window_mapcoord(wavewin, &x, &y);
	data[pos++] = (x-sim_window_basex) - sim_window_offx;
	data[pos++] = (y-sim_window_basey) - sim_window_offy;
	*datapos = pos;
}

/*
 * Routine to remove trace "tr" from the waveform window's linked list.
 */
void sim_window_removetrace(DISCHANNEL *tr)
{
	DISCHANNEL *lasttr, *otr;

	lasttr = NODISCHANNEL;
	for(otr = sim_window_plot; otr != NODISCHANNEL; otr = otr->nextdischannel)
	{
		if (otr == tr) break;
		lasttr = otr;
	}
	if (lasttr == NODISCHANNEL) sim_window_plot = tr->nextdischannel; else
		lasttr->nextdischannel = tr->nextdischannel;
}

/*
 * Routine to insert trace "tr" after "aftertr" in the waveform window's linked list.
 */
void sim_window_addtrace(DISCHANNEL *tr, DISCHANNEL *aftertr)
{
	if (aftertr == NODISCHANNEL)
	{
		tr->nextdischannel = sim_window_plot;
		sim_window_plot = tr;
	} else
	{
		tr->nextdischannel = aftertr->nextdischannel;
		aftertr->nextdischannel = tr;
	}
}

DISCHANNEL *sim_window_allocdischannel(void)
{
	REGISTER DISCHANNEL *tr;

	if (sim_window_plotfree == NODISCHANNEL)
	{
		tr = (DISCHANNEL *)emalloc(sizeof (DISCHANNEL), sim_aid->cluster);
		if (tr == 0) return(0);
	} else
	{
		tr = sim_window_plotfree;
		sim_window_plotfree = tr->nextdischannel;
	}
	tr->flags = 0;
	tr->timelimit = 0;
	tr->statelimit = 0;
	tr->valuelimit = 0;
	tr->numsteps = 0;
	tr->nameoff = 0;
	tr->buschannel = NODISCHANNEL;
	return(tr);
}

/********************************* LOW LEVEL GRAPHICS *********************************/

void sim_window_setviewport(WINDOWPART *w)
{
	float scalex, scaley;
	REGISTER INTBIG maxlines;

	w->screenlx = 0;   w->screenhx = 767;
	w->screenly = 0;   w->screenhy = 600;
	computewindowscale(w);

	scalex = (w->usehx - w->uselx) / 767.0f;
	scaley = ((w->usehy-20) - w->usely) / 600.0f;
	if (scalex > 2.0 && scaley > 2.0) sim_window_textsize = 0; else
		if (scalex > 1.0 && scaley > 1.0) sim_window_textsize = 16; else
			sim_window_textsize = 12;

	maxlines = ((w->usehy - w->usely) * 540 / 600) / 25;
	if (maxlines != sim_window_vislines)
	{
		if (sim_window_lines > sim_window_vislines)
			sim_window_vislines = sim_window_lines;
		if (sim_window_vislines > maxlines)
			sim_window_vislines = maxlines;
	}
}

/*
 * Routine to return the WINDOWPART that has the waveform display
 */
WINDOWPART *sim_window_findwaveform(void)
{
	WINDOWPART *w;

	for(w = el_topwindowpart; w != NOWINDOWPART; w = w->nextwindowpart)
		if ((w->state & WINDOWTYPE) == WAVEFORMWINDOW)
			return(w);
	return(NOWINDOWPART);
}

/*
 * Routine to return the WINDOWPART that has the schematics/layout being simulated
 */
WINDOWPART *sim_window_findschematics(void)
{
	WINDOWPART *w;
	INTBIG state;

	for(w = el_topwindowpart; w != NOWINDOWPART; w = w->nextwindowpart)
	{
		state = w->state & (WINDOWTYPE|WINDOWSIMULATING);
		if (state == (DISPWINDOW|WINDOWSIMULATING) ||
			state == (DISP3DWINDOW|WINDOWSIMULATING))
				return(w);
	}
	return(NOWINDOWPART);
}

void sim_window_mapcoord(WINDOWPART *w, INTBIG *x, INTBIG *y)
{
	REGISTER INTBIG newx, newy;

	newx = w->uselx + (INTSML)applyxscale(w, *x - w->screenlx);
	newy = w->usely + (INTSML)((*y - w->screenly) *
		((w->usehy-20) - (w->usely)) / (w->screenhy - w->screenly));
	*x = newx;   *y = newy;
}

void sim_window_scaletowindow(WINDOWPART *w, INTBIG *x, INTBIG *y)
{
	*x = muldiv(*x - w->uselx, w->screenhx - w->screenlx,
		w->usehx - w->uselx) + w->screenlx;
	*y = muldiv(*y - w->usely, w->screenhy - w->screenly,
		(w->usehy-20) - w->usely) + w->screenly;
}

/* Some Control Sequences for the graphics terminal */
void sim_window_setcolor(INTBIG color)
{
	sim_window_desc.col = (INTSML)color;
}

void sim_window_setmask(INTSML mask)
{
	sim_window_desc.bits = mask;
}

/* Graphics Commands */
void sim_window_moveto(INTBIG x, INTBIG y)
{
	sim_window_curx = x;
	sim_window_cury = y;
}

void sim_window_drawto(WINDOWPART *w, INTBIG x, INTBIG y)
{
	INTBIG fx, fy, tx, ty, cfx, cfy, ctx, cty;

	fx = sim_window_curx;   fy = sim_window_cury;
	tx = x;   ty = y;
	sim_window_mapcoord(w, &fx, &fy);
	sim_window_mapcoord(w, &tx, &ty);
	cfx = fx;   cfy = fy;
	ctx = tx;   cty = ty;
	if (clipline(&cfx, &cfy, &ctx, &cty, w->uselx, w->usehx, w->usely, w->usehy) == 0)
	{
		fx = cfx;   fy = cfy;
		tx = ctx;   ty = cty;
		screendrawline(w, (INTSML)fx, (INTSML)fy, (INTSML)tx, (INTSML)ty, &sim_window_desc, 0);
	}
	sim_window_curx = x;
	sim_window_cury = y;
}

void sim_window_drawbox(WINDOWPART *w, INTBIG xll, INTBIG yll, INTBIG xur, INTBIG yur)
{
	INTBIG lx, ly, hx, hy, swap;

	lx = xll;   ly = yll;
	hx = xur;   hy = yur;
	sim_window_mapcoord(w, &lx, &ly);
	sim_window_mapcoord(w, &hx, &hy);
	if (lx > hx) { swap = lx;   lx = hx;   hx = swap; }
	if (ly > hy) { swap = ly;   ly = hy;   hy = swap; }
	screendrawbox(w, (INTSML)lx, (INTSML)hx, (INTSML)ly, (INTSML)hy, &sim_window_desc);
}

void sim_window_drawhcstring(WINDOWPART *w, char *s)
{
	INTSML x, y;
	INTBIG px, py;

	px = sim_window_curx;   py = sim_window_cury;
	sim_window_mapcoord(w, &px, &py);
	screensettextsize(w, sim_window_textsize);
	screengettextsize(w, s, &x, &y);
	screendrawtext(w, (INTSML)px, (INTSML)(py-y/2), s, &sim_window_desc);
}

void sim_window_drawcstring(WINDOWPART *w, char *s)
{
	INTSML wid, hei;
	INTBIG px, py;

	px = sim_window_curx;   py = sim_window_cury;
	sim_window_mapcoord(w, &px, &py);
	screensettextsize(w, sim_window_textsize);
	screengettextsize(w, s, &wid, &hei);
	screendrawtext(w, (INTSML)(px-wid/2), (INTSML)py, s, &sim_window_desc);
}

void sim_window_drawulstring(WINDOWPART *w, char *s)
{
	INTSML wid, hei;
	INTBIG px, py;

	px = sim_window_curx;   py = sim_window_cury;
	sim_window_mapcoord(w, &px, &py);
	screensettextsize(w, sim_window_textsize);
	screengettextsize(w, s, &wid, &hei);
	screendrawtext(w, (INTSML)px, (INTSML)(py-hei), s, &sim_window_desc);
}

void sim_window_drawrstring(WINDOWPART *w, char *s)
{
	INTSML x, y;
	INTBIG px, py;

	screensettextsize(w, sim_window_textsize);
	screengettextsize(w, s, &x, &y);
	px = sim_window_curx;   py = sim_window_cury;
	sim_window_mapcoord(w, &px, &py);
	screendrawtext(w, (INTSML)(px-x), (INTSML)py, s, &sim_window_desc);
}

/*
 * Name: sim_windowconvertengineeringnotation
 *
 * Description:
 *	This procedure converts a floating point number and represents time
 * in engineering units such as pico, micro, milli, etc.
 *
 * Calling Arguments:
 *	time = floating point value to be converted to Engineering notation
 *	s1   = pointer to a string that will be used to print the converted value
 */
void sim_windowconvertengineeringnotation(float time, char *s1)
{
	if (time < 10.0e-15 || time >= 9999.5)
	{
		(void)sprintf(s1, "%gsec", time);
		return;
	}
	if (time >= 10.0e-15 && time < 9999.5e-15)
	{
		time = (time / 1.0e-15f) + 0.5f;
		(void)sprintf(s1, "%ldfsec", (INTBIG)time);
		return;
	}
	if (time >= 1.0e-12 && time < 9999.5e-12)
	{
		time = (time / 1.0e-12f) + 0.5f;
		(void)sprintf(s1, "%ldpsec", (INTBIG)time);
		return;
	}
	if (time >= 1.0e-9 && time < 9999.5e-9)
	{
		time = (time / 1.0e-9f) + 0.5f;
		(void)sprintf(s1, "%ldnsec", (INTBIG)time);
		return;
	}
	if (time >= 1.0e-6 && time < 9999.5e-6)
	{
		time = (time / 1.0e-6f) + 0.5f;
		(void)sprintf(s1, "%ldusec", (INTBIG)time);
		return;
	}
	if (time >= 1.0e-3 && time < 9999.5e-3)
	{
		time = (time / 1.0e-3f) + 0.5f;
		(void)sprintf(s1, "%ldmsec", (INTBIG)time);
		return;
	}
	if (time >= 1.0 && time < 9999.5)
	{
		time = time + 0.5f;
		(void)sprintf(s1, "%ldsec", (INTBIG)time);
		return;
	}
}

#endif  /* SIMAID - at top */
